import os
#import collections
from netaddr import IPNetwork, IPAddress
import vici
from vici.exception import SessionException, CommandException, EventUnknownException
from enum import IntEnum
from OpenSSL import crypto
import sys
import syslog
import socket




class IPSecIfErr(Exception):
    def __init__(self, message, code):
        super(IPSecIfErr, self).__init__(message)
        self.code = code

class IKE(IntEnum):
    V1=1
    V2=2

class IKE1_MODE(IntEnum):
    MAIN=1
    AGGRESIVE=2


ENCR_ALG = dict( [ ('des','des'), ('3des','3des'), ('aes128','aes128'), ('aes192','aes192'), ('aes256','aes256') ] )
INTEGR_ALG = dict( [ ('md5','md5'), ('sha1','sha1'), ('sha256','sha256'), ('sha384','sha384'), ('sha512','sha512') ] )
DH_GROUP = dict([ (1,'modp768'), (2,'modp1024'), (5,'modp1536'), (14,'modp2048'), (15,'modp3072'), (16,'modp4096'), (17,'modp6144'),
                  (18,'modp8192'), (22,'modp1024s160'), (23,'modp2048s224'), (24,'modp2048s256'), (25,'ecp192'), (26,'ecp224'),
                  (19,'ecp256'), (20,'ecp384'), (21,'ecp521'), (27,'ecp224bp'), (28,'ecp256bp'), (29,'ecp384bp'), (30,'ecp512bp'), (31,'curve25519') ])
BASE_MARK = 42
DEF_SCRIPT_NAME='updownipsec.sh'
SA_ENCR_ALG = dict( [ ('DES_CBC','des'), ('3DES_CBC','3des'), ('AES_CBC','aes') ] )
SA_INTEGR_ALG = dict( [ ('HMAC_MD5_96','md5'), ('HMAC_SHA1_96','sha1'), ('HMAC_SHA2_256_128','sha256'),
                        ('HMAC_SHA2_384_192','sha384'), ('HMAC_SHA2_512_256','sha512') ] )

class IPSecProposal(object):
    def __init__(self, encrypt, integrity ):
        self.encryption = encrypt
        self.integrity = integrity

    @property
    def encryption(self):
        return self._encr

    @encryption.setter
    def encryption(self, encrypt):
        if isinstance( encrypt, str ) and encrypt in list(ENCR_ALG.keys()):
            self._encr = encrypt
        else:
            raise IPSecIfErr( 'bad param', int(-1) )

    @property
    def integrity(self):
        return self._integr

    @integrity.setter
    def integrity(self, integr):
        if isinstance( integr, str ) and integr in list(INTEGR_ALG.keys()):
            self._integr = integr
        else:
            raise IPSecIfErr( 'bad param', int(-1) )

    def __eq__(self, other):
        return self.encryption == other.encryption and self.integrity == other.integrity

class AUTH_METHOD(IntEnum):
    PSK=1
    PUBKEY=2
    XAUTH=3
    EAP=4

class SYNC_STATE(IntEnum):
    NO_LOAD=1
    LOAD=2
    ACTIVE=3

class AUTH_SIDE(IntEnum):
    LOCAL=1
    REMOTE=2

class ENCODE_FORMAT(IntEnum):
    PEM=1
    DER=2

class IKE_PHASE(IntEnum):
    PHASE1=0
    PHASE2=1

class MODE_CONFIG(IntEnum):
    PUSH=1
    PULL=2

class CHILD_MODE(IntEnum):
    TRANSPORT=1
    TUNNEL=2

class IPSecAuth(object):
    def __init__(self, method=None, psk=None, crt=None, ca_crt=None, key=None, passphrase=None, user=None, psw=None):
        self._method = None
        if method:
            if isinstance(method, AUTH_METHOD):
                self._method = method
            else:
                raise IPSecIfErr('bad type of param method', int(-1))

        # if self._method == AUTH_METHOD.PUBKEY and (crt is None or ca_crt is None or key is None):
        #     raise IPSecIfErr( 'Auth. method is pubkey, need more parameters', int( -1 ) )

        self._psk = psk
        self._passphrase = passphrase
        self._crt = crt
        self._ca = ca_crt
        self._key = key
        self._user = user
        self._psw = psw

        self._crt_obj = None
        self._ca_obj = None
        self._key_obj = None

        if self._method == AUTH_METHOD.PUBKEY or self._method == AUTH_METHOD.EAP:
            if self._ca is not None:
                self.__load_ca()

            if self._crt is not None:
                self.__load_crt()

            if self._key is not None:
                # and self._passphrase is not None:
                self.__load_key()


        # if self._method == AUTH_METHOD.CERT:
        #     fname = os.path.basename(file_path)
        #     fname.split()
        #     _, f_ext = os.path.splitext(file_path)

        #     if f_ext.lower() in ('.crt'):
        #         ftype = crypto.FILETYPE_PEM
        #     else:
        #         ftype = crypto.FILETYPE_ASN1

        #     self._ssl = crypto.load_certificate(ftype, open(file_path, 'rb').read())

    def __get_fpath(self, fname):
        if '' == os.path.dirname(fname):
            return '/home/admin/user_files/{}'.format(fname)
        else:
            return fname

    def __get_ftype(self, fp):
        f = open(fp,'rb')
        line = f.readline()
        f.close()

        # Checks the binary of --BEGIN to match the binary readline
        if b'-----BEGIN' in line:
            return crypto.FILETYPE_PEM

        return crypto.FILETYPE_ASN1

    def __load_crt(self):
        try:
            if self.__get_ftype(self.__get_fpath(self._crt)) == crypto.FILETYPE_PEM:
                fpath = open(self.__get_fpath(self._crt), 'r').read()
                self._crt_obj = crypto.load_certificate(crypto.FILETYPE_PEM , fpath)
            else:
                fpath = open( self.__get_fpath(self._crt), 'rb').read()
                self._crt_obj = crypto.load_certificate(crypto.FILETYPE_ASN1 ,fpath)
        except Exception as e:
            return False
        return True

    def __load_ca(self):
        try:
            if self.__get_ftype(self.__get_fpath(self._ca)) == crypto.FILETYPE_PEM:
                fpath = open(self.__get_fpath(self._ca), 'r').read()
                self._ca_obj = crypto.load_certificate(crypto.FILETYPE_PEM, fpath)
            else:
                fpath = open(self.__get_fpath(self._ca), 'rb').read()
                self._ca_obj = crypto.load_certificate(crypto.FILETYPE_ASN1 , fpath)
        except Exception as e:
            return False
        return True

    def __load_key(self):
        try:
            if self.__get_ftype(self.__get_fpath(self._key)) == crypto.FILETYPE_PEM:
                fpath = open(self.__get_fpath(self._key),'r').read()
                keypass = self._passphrase
                if keypass is not None:
                    keypass = keypass.encode('ascii')
                self._key_obj = crypto.load_privatekey(crypto.FILETYPE_PEM, fpath, passphrase=keypass)
            else:
                syslog.syslog(syslog.LOG_INFO, "WARNING: OPENSSL CANNOT LOAD A KEY PASSPHRASE FROM BINARY KEY: ")
                fpath = open(self.__get_fpath(self._key),'rb').read()
                self._key_obj = crypto.load_privatekey(crypto.FILETYPE_ASN1, fpath, passphrase=self._passphrase)
        except Exception as e:
            return False
        return True

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, new):
        if isinstance(new, AUTH_METHOD):
            self._method = new
        else:
            raise IPSecIfErr('bad type of param method', int(-1))

    @property
    def psk(self):
        return self._psk

    @psk.setter
    def psk(self, phrase):
        self._psk = phrase

    @property
    def crt(self):
        return self._crt

    @crt.setter
    def crt(self, new):
        self._crt = new
        self._crt_obj = None
        self.__load_crt()

    @property
    def ca(self):
        return self._ca

    @ca.setter
    def ca(self, new):
        self._ca = new
        self._ca_obj=None

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, new):
        self._key=new
        self._key_obj = None
        #if self._passphrase is not None:
        self.__load_key()

    @property
    def passphrase(self):
        return self._passphrase

    @passphrase.setter
    def passphrase(self, new):
        self._passphrase = new
        if self._key is not None:
            self.__load_key()

    def get_crt_data(self, enc):
        if self._crt_obj:
            if not isinstance( enc, ENCODE_FORMAT):
                raise IPSecIfErr('Bad param type', int(-1))

            return crypto.dump_certificate(crypto.FILETYPE_PEM if enc == ENCODE_FORMAT.PEM else crypto.FILETYPE_ASN1, self._crt_obj)

        return None

    def get_ca_data(self, enc):
        if self._ca_obj:
            if not isinstance( enc, ENCODE_FORMAT):
                raise IPSecIfErr('Bad param type', int(-1))

            return crypto.dump_certificate(crypto.FILETYPE_PEM if enc == ENCODE_FORMAT.PEM else crypto.FILETYPE_ASN1, self._ca_obj)

        return None

    def get_key_data(self, enc):
        if self._key_obj:
            if not isinstance(enc, ENCODE_FORMAT):
                raise IPSecIfErr('Bad param type', int(-1))

            return crypto.dump_privatekey(crypto.FILETYPE_PEM if enc == ENCODE_FORMAT.PEM else crypto.FILETYPE_ASN1, self._key_obj)

        return None

    def get_id_from_key(self):
        comp_str = []
        if self._crt_obj:
            name_obj = self._crt_obj.get_subject()
            for (name,val) in name_obj.get_components():
                try:
                    name = name.decode()
                except (UnicodeDecodeError, AttributeError):
                    pass
                try:
                    val = val.decode()
                except (UnicodeDecodeError, AttributeError):
                    pass

                comp_str.append('{name}={value}'.format(name=name, value=val))

        return ', '.join(comp_str)

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self,new):
        self._user=new

    @property
    def psw(self):
        return self._psw

    @psw.setter
    def psw(self, new):
        self._psw = new

class ipsec_vpn_if(object):

    def __init__(self, if_name, gateway=None, ike=IKE.V2):
        self._gateway = gateway
        self._if_name = if_name
        self._local_id = None
        self._remote_id = None
        self._local_ts = None
        self._remote_ts = None
        self._ike_ver = ike
        self._ike1_mode = IKE1_MODE.MAIN
        self._mode_config = MODE_CONFIG.PULL
        self._session = vici.Session()
        self._proposals = ([],[])
        self._dh_groups = ([],[])
        self._child_mode = CHILD_MODE.TUNNEL
        #self._auth = ((None, None),(None, None))
        #auth format {(SIDE,routnd):Auth}
        self._auth = {}
        self._updown_scr = '{moddir}/{scr_name}'.format(moddir=os.path.dirname(sys.modules[__name__].__file__),scr_name=DEF_SCRIPT_NAME)
        self._vips = ['0.0.0.0', '::']

        self._sec_state = SYNC_STATE.NO_LOAD
        self._conn_state = SYNC_STATE.NO_LOAD


    def __new_conn_state(self, state):
        pass
    # if self._conn_state == SYNC_STATE.SYNC and (state == SYNC_STATE.NO_SYNC or state==SYNC_STATE.NO_LOAD):
    #     self._conn_state = state

        # if self._conn_state==SYNC_STATE.NO_LOAD and state==SYNC_STATE.SYNC:
        #     self._conn_state=state

    @property
    def name(self):
        return self._if_name

    @property
    def gateway(self):
        return self._gateway

    @gateway.setter
    def gateway(self, gateway):
        self._gateway = gateway
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    @property
    def local_id(self):
        return self._local_id

    @local_id.setter
    def local_id(self, id ):
        self._local_id = id
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    @local_id.deleter
    def local_id(self):
        self._local_id = None


    @property
    def remote_id(self):
        return self._remote_id

    @remote_id.setter
    def remote_id(self, id ):
        self._remote_id = id
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    @remote_id.deleter
    def remote_id(self):
        self._remote_id = None

    @property
    def local_ts(self):
        return self._local_ts

    @local_ts.setter
    def local_ts( self, preffix ):
        '''Set local_ts
        prefix str
        '''
        self._local_ts = IPNetwork( preffix )
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    @local_ts.deleter
    def local_ts(self):
        self._local_ts = None

    @property
    def remote_ts(self):
        ''' Rerurn remote_ts as IPNetwork object
        '''
        return self._remote_ts

    @remote_ts.setter
    def remote_ts( self, preffix ):
        '''Set remote_ts
        prefix str
        '''
        self._remote_ts = IPNetwork( preffix )
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    @remote_ts.deleter
    def remote_ts(self):
        self._remote_ts = None

    @property
    def ike(self):
        return self._ike_ver

    @ike.setter
    def ike(self, version):
        if not isinstance( version, IKE):
            raise IPSecIfErr('Bad type of version', int(-1) )

        self._ike_ver = version
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    @property
    def ike1_mode(self):
        return self._ike1_mode

    @ike1_mode.setter
    def ike1_mode(self, mode):
        if not isinstance( mode, IKE1_MODE ):
            raise IPSecIfErr('Bad type of IKE1 mode', int(-1))

        self._ike1_mode = mode
        # self.__new_conn_state(SYNC_STATE.NO_SYNC)

    def proposals_get(self, phase):
        '''
        return list of proposal obj
        '''
        if isinstance(phase, IKE_PHASE):
            return self._proposals[phase]
        return None

    def proposal_add(self, encr, integr, phase):
        if isinstance(phase, IKE_PHASE ):
            new_prop = IPSecProposal( encr, integr )
            if new_prop in self._proposals[phase]:
                return

            self._proposals[phase].append( new_prop )
            # self.__new_conn_state(SYNC_STATE.NO_SYNC)
        else:
            raise TypeError()

    def proposal_del(self, encr, integr, phase):
        if isinstance(phase, IKE_PHASE):
            del_prop = IPSecProposal( encr, integr )
            if del_prop not in self._proposals[phase]:
                return

            self._proposals[phase].remove(del_prop)
            # self.__new_conn_state(SYNC_STATE.NO_SYNC)
        else:
            raise TypeError()


    def dh_groups_get(self, phase):
        if isinstance(phase, IKE_PHASE):
            return self._dh_groups[phase]
        else:
            raise TypeError()

    def dh_group_add(self, dh_group, phase):
        if isinstance(phase, IKE_PHASE):
            if not isinstance(dh_group, int) or dh_group not in list(DH_GROUP.keys()):
                raise IPSecIfErr('bad param error', int(-1))

            if dh_group not in self._dh_groups[phase]:
                self._dh_groups[phase].append(dh_group)
                # self.__new_conn_state(SYNC_STATE.NO_SYNC)
        else:
            raise TypeError()

    def dh_group_del(self, dh_group, phase):
        if isinstance(phase, IKE_PHASE):
            if not isinstance(dh_group, int) or dh_group not in list(DH_GROUP.keys()):
                raise IPSecIfErr('bad param error', int(-1))

            if dh_group in self._dh_groups[phase]:
                self._dh_groups[phase].remove(dh_group)
                # self.__new_conn_state(SYNC_STATE.NO_SYNC)
        else:
            raise TypeError()

    def auth_add(self, side, rnd, auth):
        if isinstance(side, AUTH_SIDE) and isinstance(rnd, int) and isinstance(auth, IPSecAuth):
            self._auth[(side,rnd)]=auth
        else:
            raise IPSecIfErr( 'Bad param type error', int(-1) )

    def auth_get(self, side, rnd):
        return self._auth.get((side,rnd), None)

    def auth_del(self, side, rnd):
        if isinstance(side, AUTH_SIDE) and isinstance(rnd, int):
            self._auth[(side,rnd)]=None

    @property
    def rt_script(self):
        return self._updown_scr

    @rt_script.setter
    def rt_script(self, full_p_scr):
        self._updown_scr=full_p_scr

    @property
    def mode_config(self):
        return self._mode_config

    @mode_config.setter
    def mode_config(self, new):
        if isinstance(new, MODE_CONFIG):
            self._mode_config = new
        else:
            raise TypeError

    @property
    def child_mode(self):
        return self._child_mode

    @child_mode.setter
    def child_mode(self, new):
        if isinstance(new, CHILD_MODE):
            self._child_mode=new
        else:
            raise TypeError

    @property
    def vips(self):
        return self._vips

    @vips.setter
    def vips(self, new):
        if new is None:
            self._vips = None
            return

        if isinstance(new, str):
            try:
                ipaddr = IPAddress(new)
            except:
                raise ValueError
        elif isinstance(new, IPAddress):
            ipaddr = new
        else:
            raise TypeError

        self._vips = [str(ipaddr)]

    def __load_security(self):
        ret = True
        #ind = 0

        for ((side,rnd), auth) in self._auth.items():
            if auth is None:
                continue

            if auth.method == AUTH_METHOD.PSK:
                shared = {'id':None, 'type':None, 'data':None, 'owners':[] }
                shared['id'] = self.__get_security_id(side, rnd, auth)
                shared['type']='IKE'
                shared['data']=auth.psk
                shared['owners']=[]
                try:
                    syslog.syslog(syslog.LOG_DEBUG, 'Load PSK - {}'.format(shared))
                    self._session.load_shared(shared)
                    ret = True if ret else False
                except SessionException as s_ex:
                    syslog.syslog(syslog.LOG_ERR, '{}'.format(s_ex))
                    ret = False
                except CommandException as c_ex:
                    syslog.syslog(syslog.LOG_ERR, '{}'.format(c_ex))
                    ret = False
            elif auth.method == AUTH_METHOD.PUBKEY or auth.method == AUTH_METHOD.EAP:
                if auth.method == AUTH_METHOD.EAP:
                    shared = {'id':None, 'type':None, 'data':None, 'owners':[] }
                    shared['id']=auth.user
                    shared['type']='EAP'
                    shared['data']=auth.psw
                    try:
                        syslog.syslog(syslog.LOG_DEBUG, 'Load EAP - {}'.format(shared))
                        self._session.load_shared(shared)
                        ret = True if ret else False
                    except SessionException as s_ex:
                        syslog.syslog(syslog.LOG_ERR, '{}'.format(s_ex))
                        ret = False
                    except CommandException as c_ex:
                        syslog.syslog(syslog.LOG_ERR, '{}'.format(c_ex))
                        ret = False

                try:
                    crt = {}
                    crt['type']='X509'
                    crt['flag']='NONE'
                    crt['data']=auth.get_crt_data(ENCODE_FORMAT.PEM)
                    # if crt['data'] is None:
                    #     raise IPSecIfErr('There is not CRT', int(-1))

                    ca = {}
                    ca['type']='X509'
                    ca['flag']='CA'
                    ca['data']=auth.get_ca_data(ENCODE_FORMAT.PEM)
                    # if ca['data'] is None:
                    #     raise IPSecIfErr('There is no CA', int(-1))

                    key={}
                    key['type']='rsa'
                    key['data']=auth.get_key_data(ENCODE_FORMAT.PEM)
                    # if key['data'] is None:
                    #     raise IPSecIfErr('There is no KEY', int(-1))

                    if ca['data']:
                        self._session.load_cert(ca)

                    if crt['data']:
                        self._session.load_cert(crt)

                    if key['data']:
                        self._session.load_key(key)

                except SessionException as s_ex:
                    ret = False
                except CommandException as c_ex:
                    ret = False
            elif auth.method == AUTH_METHOD.XAUTH:
                shared = {'id':None, 'type':None, 'data':None, 'owners':[] }
                shared['id'] = self.__get_security_id(side, rnd, auth)
                shared['type']='XAUTH'
                shared['data']=auth.psw
                shared['owners']=[]
                try:
                    syslog.syslog(syslog.LOG_DEBUG, 'Load XAUTH - {}'.format(shared))
                    self._session.load_shared(shared)
                    ret = True if ret else False
                except SessionException as s_ex:
                    syslog.syslog(syslog.LOG_ERR, '{}'.format(s_ex))
                    ret = False
                except CommandException as c_ex:
                    syslog.syslog(syslog.LOG_ERR, '{}'.format(c_ex))
                    ret = False
            else:
                raise NotImplementedError

        return ret

    load_security = __load_security

    def __unload_security(self):
        ret = False
        try:
            self._session.clear_creds()
            ret = True
        except SessionException as s_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(s_ex))
        except CommandException as c_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(c_ex))
        except:
            syslog.syslog(syslog.LOG_ERR, "Exception in __unload_security")

        return ret

    unload_security = __unload_security

    def __get_prop_lst(self, phase):
        if isinstance(phase, IKE_PHASE):
            prop_lst = []
            for prop in self._proposals[phase]:
                for dh_gr in self._dh_groups[phase]:
                    prop_lst.append('{encr}-{integr}-{dh}'.format( integr=prop.integrity, encr=prop.encryption, dh=DH_GROUP.get(dh_gr)))
            return prop_lst
        else:
            raise TypeError()

    def __get_conn_name(self):
        return 'conn-{ifname}'.format(ifname=self._if_name)

    def __get_child_name(self):
        return 'ch_{ifname}'.format(ifname=self._if_name)

    def __get_security_id(self, local, rnd, auth):
        if auth.method == AUTH_METHOD.PSK:
            if local == AUTH_SIDE.LOCAL:
                return self._local_id if self._local_id else socket.gethostname()
            else:
                return self._remote_id if self._remote_id else '%any'
        elif auth.method == AUTH_METHOD.XAUTH:
            return '{direct}-{rnd}'.format(direct='loc' if local == AUTH_SIDE.LOCAL else 'rem', rnd=rnd )
        return None

    def __get_conn_auth(self, local, rnd):
        auth_sect={}
        if not isinstance(local, AUTH_SIDE):
            return None

        auth = self.auth_get(local, rnd)
        if auth is None:
            return None

        auth_sect['round'] = rnd
        if auth.method == AUTH_METHOD.PSK:
            auth_sect['auth']='psk'
            auth_sect['id'] = self.__get_security_id(local, rnd, auth)
        elif auth.method == AUTH_METHOD.PUBKEY:
            auth_sect['auth']='pubkey'
            if local == AUTH_SIDE.LOCAL:
                auth_sect['id']=auth.get_id_from_key()
        elif auth.method == AUTH_METHOD.XAUTH:
            auth_sect['auth'] = 'xauth'
            auth_sect['id'] = self.__get_security_id(local, rnd, auth)
            auth_sect['xauth_id'] = auth.user
        elif auth.method == AUTH_METHOD.EAP:
            auth_sect['auth'] = 'eap'
            auth_sect['id'] = auth.get_id_from_key()
            auth_sect['eap_id'] = auth.user
        else:
            raise NotImplementedError()

        return auth_sect

    def __load_connection(self, debug=False):
        sa_params = {}
        sa_params['version']=1 if self._ike_ver==IKE.V1 else 2
        sa_params['proposals']= self.__get_prop_lst(IKE_PHASE.PHASE1)
        sa_params['local_addrs']=['0.0.0.0', '::']
        sa_params['remote_addrs']=[self._gateway]
        sa_params['keyingtries']=0

        if self._mode_config == MODE_CONFIG.PULL:
            sa_params['pull']='yes'
        else:
            sa_params['pull']='no'

        sa_params['send_cert']='always'

        if self._vips:
            if '0.0.0.0' in self._vips:
                sa_params['vips']=['0.0.0.0', '::']
            else:
                sa_params['vips']=self._vips

        sa_params['aggressive']='yes' if self._ike1_mode == IKE1_MODE.AGGRESIVE else 'no'

        sa_params['dpd_delay']=10
        sa_params['dpd_timeout']=50

        auth = self.__get_conn_auth(AUTH_SIDE.LOCAL, 1)
        if auth:
            sa_params['local-auth-r1'] = auth

        auth = self.__get_conn_auth(AUTH_SIDE.LOCAL, 2)
        if auth:
            sa_params['local-auth-r2'] = auth

        auth = self.__get_conn_auth(AUTH_SIDE.REMOTE, 1)
        if auth:
            sa_params['remote-auth-r1'] = auth

        auth = self.__get_conn_auth(AUTH_SIDE.REMOTE, 2)
        if auth:
            sa_params['remote-auth-r2'] = auth

        child = {}
        if self._remote_ts:
            child['remote_ts']=[str(self._remote_ts)]

        if self._local_ts:
            child['local_ts']=[str(self._local_ts)]

        child['policies']='yes'
        child['mark_out']=int(self._if_name.replace('vti','')) + BASE_MARK
        child['mark_in']=int(self._if_name.replace('vti','')) + BASE_MARK
        child['updown']='{updown} {arg}'.format( updown=self._updown_scr, arg=int(self._if_name.replace('vti','')))
        child['esp_proposals']=self.__get_prop_lst(IKE_PHASE.PHASE2)

        child['dpd_action']='restart'
        child['close_action']='start'

        if self._child_mode == CHILD_MODE.TUNNEL:
            child['mode']='tunnel'
        elif self._child_mode == CHILD_MODE.TRANSPORT:
            child['mode']='transport'

        sa_params['children']={}
        sa_params['children'][self.__get_child_name()]=child
        connection={}
        connection[self.__get_conn_name()] = sa_params

        try:
            if debug:
                syslog.syslog(syslog.LOG_DEBUG, 'Connection cfg - {}'.format(connection))

            self._session.load_conn(connection)
            # self._conn_state = SYNC_STATE.SYNC
            return True
        except SessionException as s_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(s_ex))
        except CommandException as c_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(c_ex))
        except:
            pass

        self._conn_state = SYNC_STATE.NO_LOAD
        return False

    load_connection=__load_connection

    def __unload_connection(self):
        connection={'name':self.__get_conn_name()}
        ret = False

        try:
            self._session.unload_conn(connection)
            ret = True
        except SessionException as s_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(s_ex))
        except CommandException as c_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(c_ex))
        except:
            syslog.syslog(syslog.LOG_ERR, "Exception in __unload_connection")

        return ret

    unload_connection = __unload_connection

    def __log_start_stop(self, messages):
        msg_prio = syslog.LOG_INFO
        msg_log = 'Unknown'
        for message in messages:
            if message['level'] == '0':
                msg_prio = syslog.LOG_INFO
            else:
                msg_prio = syslog.LOG_ERR

            if message['group'] == 'IKE':
                msg_log = 'IKE [{sa_name}] - {msg}'.format(sa_name=message['ikesa-name'], msg=message['msg'])
            else:
                msg_log = 'Unhandled group {} - {}'.format(message['group'], message['msg'])
                #syslog.syslog(syslog.LOG_INFO, message)

            syslog.syslog( msg_prio, msg_log )

    def __connect(self, loglevel=0, timeout=5000):
        # if self._sec_state != SYNC_STATE.SYNC:
        #     if not self.__load_security():
        #         return False

        #     if not self.__load_connection():
        #         return False

        conn_initiate={'child':self.__get_child_name(), 'ike':self.__get_conn_name(),
                       'timeout':timeout, 'init-limits':0, 'loglevel':loglevel}

        try:
            res = self._session.initiate(conn_initiate)
            self.__log_start_stop(res)
            # self._conn_state = SYNC_STATE.SYNC
            return True
        except SessionException as s_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(s_ex))
            pass
        except CommandException as c_ex:
            syslog.syslog(syslog.LOG_ERR, "{}".format(c_ex))
            pass
        except:
            print('Other exception')
            pass

        return False

    connect=__connect

      #ToDo

    def __disconnect(self, loglevel=0):
        #term={'child':self.__get_child_name(), 'ike':self.__get_conn_name(), 'force':'yes',
        #    'timeout':10000, 'loglevel':4}
        term={'ike':self.__get_conn_name(), 'force':'yes', 'timeout':10000, 'loglevel':loglevel}
        try:
            res = self._session.terminate(term)
            self.__log_start_stop(res)
            # self._conn_state = SYNC_STATE.SYNC
            return True
        except SessionException as s_ex:
            syslog.syslog( syslog.LOG_ERR, '{}'.format(s_ex) )
        except CommandException as c_ex:
            syslog.syslog( syslog.LOG_ERR, '{}'.format(c_ex) )
        except:
            print('Other exception')
            pass

        return False

    disconnect=__disconnect

    def __get_connection(self):
        gen = self._session.list_conns(filters={ 'name': self.__get_conn_name()})
        #gen = self._session.list_conns()
        print(gen)
        for item in gen:
            print(item)

    get_connection=__get_connection

    def __get_child_sa(self, ch_info=None):
        ch = {}
        if ch_info is not None:
            ch['chname'] = "{}[{}]".format(ch_info.get('name', None), ch_info.get('uniqueid', None))
            ch['state'] = ch_info.get('state', None)
            ch['mode'] = ch_info.get('mode', None)
            ch['proto'] = ch_info.get('protocol', None)
            ch['encap'] = ch_info.get('encap', None)
            ch['encr-alg'] = SA_ENCR_ALG.get(ch_info.get('encr-alg', None), ch_info.get('encr-alg', None))
            ch['encr-ksize'] = ch_info.get('encr-keysize', "")
            ch['integ-alg'] = SA_INTEGR_ALG.get(ch_info.get('integ-alg', None), ch_info.get('integ-alg', None))
            dh_gr_ch = ch_info.get('dh-group', None)
            ch['dh-group'] = list(DH_GROUP.keys())[
                list(DH_GROUP.values()).index(bytes(dh_gr_ch).decode().lower().replace("_", ""))] if dh_gr_ch else dh_gr_ch
            ch['bytes-in'] = ch_info.get('bytes-in', None)
            ch['bytes-out'] = ch_info.get('bytes-out', None)
            ch['pkt-in'] = ch_info.get('packets-in', None)
            ch['pkt-out'] = ch_info.get('packets-out', None)
            ch['local-ts'] = ch_info.get('local-ts', None)
            ch['remote-ts'] = ch_info.get('remote-ts', None)
        return ch

    def __get_sa(self, all_child=False):
        # {'version':0, 'state':'', local, remote, local_id, remote_id, nat-local, nat-any, encr-alg, encr-ksize, integr-alg, dh-gr, established, rekey-time, vips, childs:[]}
        # child {name, state, mode, proto, enc, encr-alg, encr-keysiz, dh-group, bytes-in, bytes-out, pkt-in, pkt-out, use-in, use-out, local-ts, remote-ts}

        try:
            gen = self._session.list_sas(filters={'ike':self.__get_conn_name()})
            #gen = self._session.list_sas(filters={'ike':'dddf'})
            sa_info = next(gen)
            if sa_info is None:
                return None
            gen.close()
        except StopIteration:
            return None

        ret = {}
        ret['conname'] = "{}[{}]".format(self.__get_conn_name(), sa_info[self.__get_conn_name()].get('uniqueid', None))
        ret['version'] = sa_info[self.__get_conn_name()].get('version', None)
        ret['state'] = sa_info[self.__get_conn_name()].get('state', None)
        ret['local']='{ip}:{port}'.format(ip=sa_info[self.__get_conn_name()].get('local-host', None), port=sa_info[self.__get_conn_name()].get('local-port', None))
        ret['remote']='{ip}:{port}'.format(ip=sa_info[self.__get_conn_name()].get('remote-host', None), port=sa_info[self.__get_conn_name()].get('remote-port', None))
        ret['local-id']=sa_info[self.__get_conn_name()].get('local-id', None)
        ret['remote-id']=sa_info[self.__get_conn_name()].get('remote-id', None)
        ret['nat-local']=sa_info[self.__get_conn_name()].get('nat-local', None)
        ret['nat-any']=sa_info[self.__get_conn_name()].get('nat-any', None)
        ret['encr-alg']=SA_ENCR_ALG.get(sa_info[self.__get_conn_name()].get('encr-alg', None), sa_info[self.__get_conn_name()].get('encr-alg', None))
        ret['encr-ksize']=sa_info[self.__get_conn_name()].get('encr-keysize', "")
        ret['integ-alg']=SA_INTEGR_ALG.get(sa_info[self.__get_conn_name()].get('integ-alg', None), sa_info[self.__get_conn_name()].get('integ-alg', None))
        dh_gr = sa_info[self.__get_conn_name()].get('dh-group', None)
        ret['dh-group']=list(DH_GROUP.keys())[list(DH_GROUP.values()).index(bytes(dh_gr).decode().lower().replace("_",""))] if dh_gr is not None else dh_gr
        ret['vips']=sa_info[self.__get_conn_name()].get('local-vips', None)[0] if sa_info[self.__get_conn_name()].get('local-vips', None) else None
        ret['established']=sa_info[self.__get_conn_name()].get('established', "-")
        ret['rekey-time']=sa_info[self.__get_conn_name()].get('rekey-time', "-")

        if not sa_info[self.__get_conn_name()].get('child-sas', None):
            ret['child']={}
            return ret

        if all_child is True:
            # In case of all child the "child" field is list of dictonaries {'child' : [dic_child1, dic_child2,...]}
            ret['child']=[]
            for ch_info in sa_info[self.__get_conn_name()]['child-sas'].items():
                ch = self.__get_child_sa(ch_info[1])
                if ch:
                    # return list of dicts
                    ret['child'].append(ch)
        else:
            # Preserve original functionality where child field is dictonay {'child': dictonary}
            ch = {}
            ch_info = dict(sa_info[self.__get_conn_name()]['child-sas'].items())
            ch = self.__get_child_sa(ch_info)
            if ch:
                # return single dict
                ret['child']=ch

        return ret

    get_sa=__get_sa

