#!/usr/bin/env python3
import os
import logging
import logging.handlers
import vppapi_if.vppapi_if as vppif
import ipaddress
import debinterface
import tempfile
import random
import subprocess
from enum import IntEnum
import sys
# Enabling import for current directory
sys.path.append(os.path.dirname(__file__))
from tuntap_ovs import TunTap

LOG_NAME = 'loop-cls'

def ip_to_bytes(ip_str):
    ip=ip_str.split("/")[0]
    ip_as_bytes = bytes(map(int, ip.split('.')))
    return ip_as_bytes

class LoopIfErr(Exception):
    def __init__(self, message, code):
        super(LoopIfErr, self).__init__(message)
        self.code = code

class LoopInterface(object):
    LOOP_OFFSET=600
    CFG_FILE_PATH='/etc/network/interfaces.d/ifcfg-{}'

    def __init__(self, name, ipv4_prefix=None, ipv6_prefix=None, enable=None, logger=None, log_level=logging.DEBUG, network_mode=None, mtu=1500):
        if enable:
            self._enable = enable
        else:
            self._enable = False
        self._ipv4 = None
        self._ipv6 = None
        self.tap_swidx = None
        self.loop_swidx = None
        self._use_drctl=True
        self._mtu = mtu

        # The defualt value of network mode is OVS - for now
        if network_mode: 
            self.network_mode = network_mode
        else: 
            self.logger.debug('NO NETWORK MODE SET')
            self.network_mode = "OVS"
        self._ovs_tap = None

        macarr =  [0xfe, random.getrandbits(8),
           random.getrandbits(8), random.getrandbits(8),
           random.getrandbits(8), random.getrandbits(8)]
        self._mac = ':'.join(["%02x" % x for x in macarr])

        self.logger = None
        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level )

        self._name = name
        self.logger.debug('LoopInterface for \'{}\' - init '.format(self._name))
        self.logger.debug('Network mode : {}'.format(self.network_mode))

        if ipv4_prefix:
            if not isinstance( ipv4_prefix, (ipaddress.IPv4Interface)):
                raise LoopIfErr('Bad type of ipv4_prefix', -1)
            self._ipv4 = ipv4_prefix
        
        if ipv6_prefix:
            if not isinstance( ipv6_prefix, (ipaddress.IPv6Interface)):
                raise LoopIfErr('Bad type of ipv6_prefix', -1)
            self._ipv6 = ipv6_prefix
        
        if enable:  
            if (self._ipv6 or self._ipv4):
                if   self.network_mode == "OVS": 
                    self._create_ovs_loopback()
                elif self.network_mode == "VPP": 
                    self._create_vpp_loopack()
                    self._enable = True
                
        if 'yes' == os.environ.get('RTCTL', ''):
            self._use_drctl=False


        #self.__update_cfg_file()
        #self.down_sys_if()
        
        #if self._enable:
        #    self.up_sys_if()
           
           
    def __del__(self):
        #interfaces = debinterface.Interfaces(interfaces_path=self.__get_cfg_fname())
        #interfaces.downAdapter(self.__get_sys_ifname())

        #os.unlink(self.__get_cfg_fname())

        self._delete_vpp_loopack()

    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)

    def __get_sys_ifname(self):
        return '{}'.format(self._name)

    def __get_cfg_fname(self):
        return self.CFG_FILE_PATH.format(self.__get_sys_ifname())

    def __get_ipv4_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv4_set = debinterface.NetworkAdapter(sys_ifname)
        ipv4_set.setName(sys_ifname)
        ipv4_set.setAddrFam('inet')

        if self._ipv4:
            ipv4_set.setAddressSource('static')
            ipv4_set.setAddress(str(self._ipv4.ip))
            ipv4_set.setNetmask(str(self._ipv4.network.netmask))
        else:
            ipv4_set.setAddressSource('manual')

        return ipv4_set
        
    def __get_ipv6_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv6_set = debinterface.NetworkAdapter(sys_ifname)
        ipv6_set.setName(sys_ifname)
        ipv6_set.setAddrFam('inet6')

        if self._ipv6:
                ipv6_set.setAddressSource('static')
                ipv6_set.setAddress(str(self._ipv6.ip))
                ipv6_set.setNetmask(str(self._ipv6.network.prefixlen))
        else:
                return None

        ipv6_set.appendPreUp('echo 0 > /proc/sys/net/ipv6/conf/{}/accept_dad'.format(sys_ifname))
        return ipv6_set

    def __update_cfg_file(self):
        sys_ifname=self.__get_sys_ifname()
        interfaces_path=self.__get_cfg_fname()

        if not os.path.exists(interfaces_path):
            with open(interfaces_path, 'w') as f:
                pass

        (fh, backup) = tempfile.mkstemp(dir='/tmp')
        os.close(fh)

        interfaces = debinterface.Interfaces(interfaces_path=interfaces_path, backup_path=backup)
        interfaces.removeAdapterByName(sys_ifname)

        ipv4_set = self.__get_ipv4_adapter_cfg()
        ipv6_set = self.__get_ipv6_adapter_cfg()

        adapter = ipv4_set if ipv4_set else ipv6_set

        adapter.setAuto(self._enable)
        adapter.setHotplug(True)

        index = 0
        if ipv4_set:
            interfaces.addAdapter( ipv4_set.export(), index)
            index += 1
        
        if ipv6_set:
            interfaces.addAdapter( ipv6_set.export(), index)

        interfaces.writeInterfaces()

        os.unlink(backup)


    @property
    def name(self):
        return self._name

        
    def _create_vpp_loopack(self):
        if not self._name:
            self.logger.error('Interface name is not set!')
            return

        self.logger.debug("Creating loopback {}" .format(self._name))
        vpp_if = vppif.VppApiInterface(LOG_NAME)

        #get tap ind
        tap_id = int(self._name.replace('lo',''))

        if self._ipv4:
            ipv4ip=self._ipv4.ip
            # ipv4prefixlen=self._ipv4._prefixlen
            host_ip4_addr_set=True
        else:
            ipv4ip=None
            # ipv4prefixlen=None
            host_ip4_addr_set=False
                        
        if self._ipv6:
            ipv6ip=self._ipv6.ip
            # ipv6prefixlen=self._ipv6._prefixlen
            host_ip6_addr_set=True
        else:
            ipv6ip=None
            # ipv6prefixlen=None
            host_ip6_addr_set=False
  
        self.logger.debug("call create_tap_interface({}, id={}, host_mac_addr={}, mac_addr={},\nhost_mac_addr_set={}, host_ip4_addr_set={}, host_ip4_addr={}".format(
                                                     self._name,  tap_id+self.LOOP_OFFSET,  self._mac,  self._mac,  
                                                     True, host_ip4_addr_set, ipv4ip))
                                                     
        (ret, tap_swidx)=vpp_if.create_tap_interface(self._name, id=tap_id+self.LOOP_OFFSET, host_mac_addr=self._mac, mac_addr=self._mac,  
                                                     host_mac_addr_set=True,
                                                     host_ip4_addr_set=host_ip4_addr_set, host_ip4_addr=self._ipv4,
                                                     host_ip6_addr_set=host_ip6_addr_set, host_ip6_addr=self._ipv6,
                                                     host_mtu_size=self._mtu)
        
        if ret:
            self.logger.error('Error creating tap if {}'.format(ret))
        self.tap_swidx = tap_swidx
        #Create bridge-domain
        ret = vpp_if.create_bridge_domain(self.tap_swidx)

        #Attach interface to bridge-domain
        vpp_if.add_to_bridge(self.tap_swidx,self.tap_swidx)
        vpp_if.set_interface_learning(self.tap_swidx, disable=1)
        
        (ret, loop_swidx)=vpp_if.create_loopback_interface(id=tap_id+self.LOOP_OFFSET, mac_addr=self._mac)
        if ret:
            self.logger.error('Error creating loop if {}'.format(ret))
        self.loop_swidx = loop_swidx
        
        if self._ipv4:
            ret = vpp_if.add_interface_address(loop_swidx, str(self._ipv4))
            if ret:
                self.logger.error('Error adding address {}'.format(ret))

        if self._ipv6:
            ret = vpp_if.add_interface_address(loop_swidx, str(self._ipv6))
            if ret:
                self.logger.error('Error adding address {}'.format(ret))

        vpp_if.add_to_bridge(self.tap_swidx, self.loop_swidx, port_type=vppif.L2_API_PORT_TYPE_BVI)
        
        vpp_if.ifup(tap_swidx)
        vpp_if.ifup(loop_swidx)
        
        if self._use_drctl:
            ret = vpp_if.enable_drctl(tap_swidx, self._mac)
            
            if ret:
                self.logger.error('Error enable routectl {}'.format(ret))
            
            if self._ipv4:    
                ret = vpp_if.drctl_set_ip(tap_swidx, ipv4ip, delete=0)
                if ret:
                    self.logger.error('Error set router address {}'.format(ret))

            if self._ipv6:    
                ret = vpp_if.drctl_set_ip(tap_swidx, ipv6ip, delete=0)
                if ret:
                    self.logger.error('Error set router address {}'.format(ret))
        else:
            ret = vpp_if.enable_routectl(tap_swidx, loop_swidx)
            if ret:
                self.logger.error('Error enable routectl {}'.format(ret))
            
            if self._ipv4:    
                ret = vpp_if.routectl_set_ip(ipv4ip, disable=0)
                if ret:
                    self.logger.error('Error set router address {}'.format(ret))

            if self._ipv6:    
                ret = vpp_if.routectl_set_ip(ipv6ip, disable=0)
                if ret:
                    self.logger.error('Error set router address {}'.format(ret))
        del(vpp_if)


    def _create_ovs_loopback(self):
        if not self._name:
            self.logger.error('Interface name is not set!')
            return
        
        self.logger.debug("Creating loopback {}".format(self._name))
        
        if self._ipv4:
            ipv4ip = self._ipv4.with_prefixlen
        else:
            ipv4ip = None

        if self._ipv6:
            ipv6ip = self._ipv6.ip.exploded
        else:
            ipv6ip = None

        self.logger.debug(f"create_tap_interface_ovs : \
            name={self._name}, mac={self._mac}, ipv4ip={ipv4ip}, ipv6ip={ipv6ip}")
        self._ovs_tap = TunTap(nic_type="Tap",nic_name=self._name, logger=self.logger)

        self._ovs_tap.config(ip4=ipv4ip, ip6=ipv6ip, mtu=self._mtu)
            
        
    def _delete_vpp_loopack(self):
        
        if not self.tap_swidx:
            return
        self.logger.debug("Deleting loopback {}" .format(self._name))

        if self._ipv4:
            ipv4ip=self._ipv4.ip
            # ipv4prefixlen=self._ipv4._prefixlen
            
        if self._ipv6:
            ipv6ip=self._ipv6.ip
            # ipv6prefixlen=self._ipv6._prefixlenq
                        
        vpp_if = vppif.VppApiInterface(LOG_NAME)

        #get tap ind
        vpp_if.ifdown(self.tap_swidx)
        vpp_if.ifdown(self.loop_swidx)
        
        if self._use_drctl:
            vpp_if.disable_drctl(self.tap_swidx, self._mac)
        else:
            vpp_if.disable_routectl(self.tap_swidx, self.loop_swidx)
        
        
        if self._ipv4:
            if self._use_drctl:
                vpp_if.drctl_set_ip(self.tap_swidx, ipv4ip, delete=1)
            else:
                vpp_if.routectl_set_ip(ipv4ip, disable=1)
            vpp_if.del_interface_address(self.loop_swidx, str(self._ipv4))
        
        if self._ipv6:
            if self._use_drctl:
                vpp_if.drctl_set_ip(self.tap_swidx, ipv6ip, delete=1)
            else:
                vpp_if.routectl_set_ip(ipv6ip, disable=1)
            vpp_if.del_interface_address(self.loop_swidx, str(self._ipv6))
        
        vpp_if.delete_loopback_interface(self.loop_swidx)
        vpp_if.delete_tap_interface(self.tap_swidx)
        
        self.tap_swidx=None
        del vpp_if


    def _delete_ovs_loopback(self):
        self.logger.debug("Deleting loopback {}".format(self._name))

        if self._ovs_tap:
            self._ovs_tap.close()
        else:
            self.logger.debug('Deleting interface (manually, ovs_tap object missing) \'{}\''.format(self._name))
            subprocess.check_call('ip tuntap delete mode tap ' + self._name , shell=True)


    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, new):
        if not isinstance(new, bool):
            raise VxLanIfErr("enable expect boolean", -1)

        self.logger.debug("Set enable from {} to {} for {}".format(self._enable, new, self._name))
        if self._enable == new:
            return

        if self._enable:
            if self.network_mode == "VPP":
                self._delete_vpp_loopack()
            elif self.network_mode == "OVS":
                self._delete_ovs_loopback()

        self._enable = new
        #self.__update_cfg_file()

        if new:
            if self._ipv4 or self._ipv6:
                if self.network_mode == "VPP":
                    self._create_vpp_loopack()
                elif self.network_mode == "OVS":
                    self._create_ovs_loopback()
        

    @property
    def ip4(self):
        return self._ipv4

    @ip4.setter
    def ip4(self, new):
        if new and not isinstance(new, ipaddress.IPv4Interface):
            raise VxLanIfErr("ipv4 expect ipaddress.IPv4Interface or None", -1)

        if self._ipv4 == new:
            return

        #if self._enable:
        #    self.down_sys_if()
        
        if self._enable:
            if self.network_mode == "VPP": self._delete_vpp_loopack()
            elif self.network_mode == "OVS": 
                self.logger.debug('Deleting interface (because no IPv4/IPv6 is set)')
                if self._ovs_tap:
                    self.logger.debug('Deleting interface (_ovs_tap closing)')
                    self._ovs_tap.close()
                else:
                    self.logger.error('Error deleting interface \'{self._name}\' - no ovs_tap object found')
        
        self._ipv4 = new
        #self.__update_cfg_file()
        if new and self._enable:
            if self.network_mode == "VPP": self._create_vpp_loopack()
            elif self.network_mode == "OVS": self._create_ovs_loopback()

        #if self._enable:
        #    self.up_sys_if()

    @property
    def ip6(self):
        return self._ipv6

    @ip6.setter
    def ip6(self, new):
        if new and not isinstance(new, ipaddress.IPv6Interface):
            raise VxLanIfErr("ipv4 expect ipaddress.IPv6Interface or None", -1)

        if self._ipv6 == new:
            return

        #if self._enable:
        #    self.down_sys_if()

        if self._enable:
            if   self.network_mode == "VPP": self._delete_vpp_loopack()
            elif self.network_mode == "OVS": 
                self.logger.debug('Deleting interface \'{}\''.format(self._name))
                if self._ovs_tap:
                    self.logger.debug('Deleting interface (ovs_tap closing)')
                    self._ovs_tap.close()
                else:
                    self.logger.error(f'Error deleting interface {self._name} - no ovs_tap object found')
        
        self._ipv6 = new
        #self.__update_cfg_file()
        
        if new and self._enable:
            if   self.network_mode == "VPP": self._create_vpp_loopack()
            elif self.network_mode == "OVS": self._create_ovs_loopback()

        #if self._enable:
        #    self.up_sys_if()

    @property
    def mtu(self):
        return self._mtu

    @mtu.setter
    def mtu(self, new):
        if self._mtu == new:
            return

        if new and self._enable:
            if   self.network_mode == "VPP":
                vpp_if = vppif.VppApiInterface(LOG_NAME)
                vpp_if.set_interface_mtu(self.loop_swidx, new )
                vpp_if.set_interface_mtu(self.tap_swidx, new )

            elif self.network_mode == "OVS":
                self._ovs_tap.config(mtu=new)

        self._mtu = new
        #self.__update_cfg_file()

    def down_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifdown", "--force", self.__get_sys_ifname()])

    def up_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifup", self.__get_sys_ifname()])
