#!/usr/bin/env python3
import os
import logging
import logging.handlers
import vppapi_if.vppapi_if as vppif
import ipaddress
import debinterface
import tempfile
from enum import IntEnum
import time
import random
import debinterface
from pyroute2 import IPDB,IPRoute,NetNS,NetlinkError
import subprocess

DEF_PPPOE_METRIC=130

class AUTH_TYPE(IntEnum):
    PAP=1
    CHAP=2

class PPPoEClientIfErr(Exception):
    def __init__(self, message, code):
        super(PPPoEClientIfErr, self).__init__(message)
        self.code = code

class PPPoEClientInterface(object):
    PPP_OFFSET=700
    PPP_DHCPD_OFFSET=800
    CFG_FILE_PATH='/etc/network/interfaces.d/ifcfg-{}'
    INTERFACES_FILE='/etc/network/interfaces'
    PPPOX_TO_HOST_IF_NAME_DIR_PATH = '/var/run/pppox-to-host-ifname-map'

    def __init__(self, name, auth_type=AUTH_TYPE.PAP, ethif=None, username=None, password=None, logger=None, log_level=logging.ERROR, enable=False,
                            idle_timeout=0, retry_timeout=0, metric=DEF_PPPOE_METRIC, ipv6=False, ac_name=None, ip_pass=False ):

        self.LOG_NAME = 'pppoe-cl-cls'
        self.logger = None
        self._pppoe_bind_idx = None
        self._pppoe_idx = None

        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level ) 

        self._name = name
        self._enable = enable
        self._ethif = None
        self._idle_timeout = idle_timeout
        self._retry_timeout = retry_timeout
        self._metric = metric
        self._mtu = 1492
        self._tapidx = None
        self._dhcpidx = None
        self._ipv6 = ipv6
        self._ac_name = ac_name
        self._ip_pass = ip_pass

        self._mac = self.__get_ippass_mac()

        if auth_type and not isinstance(auth_type, AUTH_TYPE):
            raise PPPoEClientIfErr('Bad authentication type', -1)
        self._auth_type = auth_type

        if ethif:
            if not isinstance(ethif, str):
                raise PPPoEClientIfErr('Bad underlying interface type', -1)

            self._ethif = ethif
            if self._enable:
                self._create_pppoeclient_if()
                # self.up_sys_if()
    
        if username and not isinstance(username, str):
            raise PPPoEClientIfErr('Bad username type', -1)
        self._username = username

        if password and not isinstance(password, str):
            raise PPPoEClientIfErr('Bad password type', -1)
        self._password = password

        self._create_ppp_tapif()
        self.__update_cfg_file()

        try:
            subprocess.check_call(['ip', 'link', 'set', 'dev', self.__get_sys_ifname(), 'arp', 'off'])
        except subprocess.CalledProcessError as e:
            self.logger.error('Failed to disable ARP for PPP interface {}' .format(self.__get_sys_ifname()))

        self.down_sys_if()

        # Configure host interface first in order to apply IP address when connects to the server
        self._set_pppoeclient_auth()

        if self._ip_pass == True:
            self._create_ippass_dhcpd_if()
            self.enable_ippass()
            self.enable_ippass_wanif()
        self.logger.debug('PPPoEClientInterface adding PPPoE config for \'{}\' - ethif {} user {} pass {}'.format(self._name, self._ethif, self._username, self._password))

    def __del__(self):
        self.logger.debug("PPPoEClientInterface {} __del__".format(self._name))
        # self._delete_ppp_tapif()

        os.unlink(self.__get_cfg_fname())

        if self._ethif and self._pppoe_bind_idx and self._enable:
            self._delete_pppoeclient_if()
            self.logger.debug("PPPoEClientInterface {} deleted".format(self._name))

        self._delete_ppp_tapif()
        self._delete_ippass_dhcpd_if()
        self._delete_ns()

        if self._ip_pass:
            self.disable_ippass()

    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=self.LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)

    @property
    def name(self):
        return self._name

    @property
    def auth_type(self):
        return self._auth_type

    @auth_type.setter
    def auth_type(self, new):
        if new and not isinstance(new, AUTH_TYPE):
            raise PPPoEClientIfErr('Bad authentication type param', -1)

        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)
 
        self._auth_type = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, new):
        if new and not isinstance(new, str):
            raise PPPoEClientIfErr('Bad username type param', -1)
        
        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)
        
        self._username = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, new):
        if new and not isinstance(new, str):
            raise PPPoEClientIfErr('Bad password type param', -1)
        
        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)
        
        self._password = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    @property
    def ethif(self):
        return self._ethif

    def _create_pppoeclient_if(self):
        if not self._ethif:
            return

        idx = None
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        #get bd by interface
        ph_if_idx = vpp_if.get_ifidx_by_name(self._ethif)
        bd_idx = vpp_if.get_interface_bd_id(ph_if_idx)
        self._pppoe_bind_idx = vpp_if.get_bd_bvi_by_bdidx(bd_idx)
        if self._pppoe_bind_idx == 4294967295:
            self._pppoe_bind_idx = None
            return

        (ret, idx) = vpp_if.pppoeclient_add_client(eth_ifidx=self._pppoe_bind_idx, 
                                                   ac_name=0 if self._ac_name == None else self._ac_name,
                                                   idle_timeout=self._idle_timeout, 
                                                   retry_timeout=self._retry_timeout) #, host_uniq=1111
        self.logger.debug("vpp pppoeclient return {} index is {}".format(ret, idx))
        if ret == 0:
            self._pppoe_idx = idx
            self.__update_mapping_file(idx)

        if self._tapidx:
            vpp_if.pppoeclient_set_host(ppp_ifidx=self._pppoe_idx, host_idx=self._tapidx, metric=self._metric)
        del(vpp_if)

    def _delete_pppoeclient_if(self):
        if not self._pppoe_bind_idx:
            return
        
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        vpp_if.pppoeclient_del_client(self._pppoe_bind_idx)

        if self._tapidx and self._pppoe_idx:
            vpp_if.pppoeclient_set_host(ppp_ifidx=self._pppoe_idx, host_idx=self._tapidx, metric=self._metric, is_add=False)
        self.__update_mapping_file(self._pppoe_idx, delete=1)
        self._pppoe_bind_idx = None
        del(vpp_if)
        self.disable_ippass_wanif()
        self._pppoe_idx = None

    def _set_pppoeclient_auth(self):
        if not self._pppoe_idx:
            self.logger.error("PPPoE interface do not exists")
            return

        if not self._username or not self._password:
            self.logger.error('Username of password is not set')
            return

        if not self._auth_type:
            self.logger.error('Authentication method is not set')
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        ret = vpp_if.pppoeclient_set_auth(self._pppoe_idx, self._username, self._password, self._auth_type.value)
        self.logger.debug('pppoeclient_set_auth ret {}'.format(ret))
        del(vpp_if)

    @ethif.setter
    def ethif(self, new):
        if new and not isinstance(new, str):
            raise PPPoEClientIfErr('Bad ethif type param', -1)

        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)
 
        self._ethif = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, new):
        if new and not isinstance(new, bool):
            raise PPPoEClientIfErr('Bad enable type param', -1)

        if self._enable == new:
            return

        if self._enable:
            self._delete_pppoeclient_if()
            # self.down_sys_if()
        else:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()
            # self.up_sys_if()

        self._enable = new

    @property
    def idle_timeout(self):
        return self._idle_timeout

    @idle_timeout.setter
    def idle_timeout(self, new):
        if new and not isinstance(new, int):
            raise PPPoEClientIfErr('Bad idle_timeout type param', -1)

        if self._idle_timeout == new:
            return
        
        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)

        self._idle_timeout = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    @property
    def retry_timeout(self):
        return self._retry_timeout

    @retry_timeout.setter
    def retry_timeout(self, new):
        if new and not isinstance(new, int):
            raise PPPoEClientIfErr('Bad retry_timeout type param', -1)

        if self._retry_timeout == new:
            return
        
        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)

        self._retry_timeout = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    @property
    def metric(self):
        return self._metric

    @metric.setter
    def metric(self, new):
        if new and not isinstance(new, int):
            raise PPPoEClientIfErr('Bad metric type param', -1)

        if self._metric == new:
            return
        
        self._metric = new

        self.down_sys_if()
        if self._enable is True:
            self.up_sys_if()

        if self._pppoe_idx and self._tapidx:
            vpp_if = vppif.VppApiInterface(self.LOG_NAME)
            vpp_if.pppoeclient_set_host(ppp_ifidx=self._pppoe_idx, host_idx=self._tapidx, metric=self._metric)
            del(vpp_if)

    @property
    def mtu(self):
        return self._mtu

    @mtu.setter
    def mtu(self, new):
        if new and not isinstance(new, int):
            raise PPPoEClientIfErr('Bad MTU type param', -1)
        self._mtu = new
        self._set_interface_mtu()

    def _set_interface_mtu(self):
        """ Need to configure MTU on both 
            - pppoxX interface in VPP
            - host repersentation in user space
        """
        if not self._mtu:
            self.logger.error('MTU is not configured')
            return

        with IPRoute() as ipr:
            index = ipr.link_lookup(ifname=self.__get_sys_ifname())[0]
            ipr.link('set', index=index, mtu=self._mtu)

        if self._pppoe_idx:
            vpp_if = vppif.VppApiInterface(self.LOG_NAME)
            ret = vpp_if.set_interface_mtu(self._pppoe_idx, self._mtu)
            self.logger.debug('set_interface_mtu ret {}'.format(ret))
            del(vpp_if)

    @property
    def ipv6(self):
        return self._ipv6

    @ipv6.setter
    def ipv6(self, new):
        if new and not isinstance(new, bool):
            raise PPPoEClientIfErr('Bad ipv6 type param', -1)

        if self._ipv6 == new:
            return

        self._ipv6 = new

    @property
    def ac_name(self):
        return self._ac_name

    @ac_name.setter
    def ac_name(self, new):
        if new and not isinstance(new, str):
            raise PPPoEClientIfErr('Bad AC-name type param', -1)

        if self._ac_name == new:
            return
        
        if self._ethif and self._pppoe_bind_idx:
            self._delete_pppoeclient_if()
            time.sleep(10)
 
        self._ac_name = new
        if self._enable:
            self._create_pppoeclient_if()
            self._set_pppoeclient_auth()
            self.enable_ippass_wanif()

    def _create_ppp_tapif(self):
        if not self._name:
            self.logger.error('Interface name is not set!')
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        #get tap ind
        tap_id = int(self._name.replace('ppp',''))
        self.logger.debug('ppp ID {}'.format(tap_id))

        (ret, tapidx)=vpp_if.create_tap_interface(self.__get_sys_ifname(), id=tap_id+self.PPP_OFFSET, 
                                                            host_mac_addr=self._mac, 
                                                            mac_addr=self._mac,
                                                            host_mac_addr_set=True,
                                                            host_mtu_size=self._mtu)

        if 0 != ret:
            self.logger.error('create_tap_interface error - {}' .format(ret))
            return

        #Create bridge-domain
        ret = vpp_if.create_bridge_domain(tapidx)
        #Attach interface to bridge-domain
        vpp_if.add_to_bridge(tapidx,tapidx)
        # configure drctl and l2 emulation
        vpp_if.set_interface_learning(tapidx, disable=1)
        # setting l2 emulation on interface returns always "unpack_from requires a buffer of at least 18 bytes for unpacking 8 bytes at offset 10 (actual buffer size is 10)"
        # however it does set the l2 emulation correctly, therefore using try/except to avoid failure due to non 0 value
        try:
            ret = vpp_if.l2_emulation_enable_disable(tapidx, enable=1)
        except Exception as e:
            self.logger.info(f"L2 emulation set on {tapidx}. Disregard exception: {e}")
        
        vpp_if.enable_drctl(tapidx, mac=self._mac)
        vpp_if.add_del_vrf_table(table_id=tap_id+self.PPP_OFFSET)
        vpp_if.set_interface_vrf(idx=tapidx, table_id=tap_id+self.PPP_OFFSET)
        vpp_if.ifup(tapidx)
        if self._pppoe_idx:
            vpp_if.pppoeclient_set_host(ppp_ifidx=self._pppoe_idx, host_idx=tapidx, metric=self._metric)

        self._tapidx = tapidx
        del(vpp_if)

    def _delete_ppp_tapif(self):
        if not self._tapidx:
            self.logger.error('Interface does not exists!')
            return
        
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        if self._pppoe_idx:
            vpp_if.pppoeclient_set_host(ppp_ifidx=self._pppoe_idx, host_idx=self._tapidx, metric=self._metric, is_add=False)
        vpp_if.ifdown(self._tapidx)
        vpp_if.disable_drctl(self._tapidx, self._mac)
        # deleting l2 emulation from interface returns always "unpack_from requires a buffer of at least 18 bytes for unpacking 8 bytes at offset 10 (actual buffer size is 10)"
        # however it does delete the l2 emulation correctly, therefore using try/except to avoid failure due to non 0 value
        try:
            ret = vpp_if.l2_emulation_enable_disable(self._tapidx, enable=0)
        except Exception as e:
            self.logger.info(f"L2 emulation deleted from {self._tapidx}. Disregard exception: {e}")
        
        vpp_if.delete_tap_interface(self._tapidx)
        #vpp_if.set_interface_vrf(idx=self._tapidx, table_id=0)
        vpp_if.delete_tap_interface(self._tapidx)
        tap_id = int(self._name.replace('ppp',''))
        #ret = vpp_if.add_del_vrf_table(table_id=tap_id+self.PPP_OFFSET, is_add=False)
        # cleanup ARP term entries
        vpp_if.flush_bd_ip_mac_entries(bd_id=self._tapidx)

        self._tapidx = None
        del(vpp_if)

    def __get_sys_ifname(self):
        # No 'br-' prefix
        return self._name

    def __get_cfg_fname(self):
        return self.CFG_FILE_PATH.format(self.__get_sys_ifname())

    def down_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifdown", "--force", self.__get_sys_ifname()])

    def up_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifup", self.__get_sys_ifname()])

    def __update_cfg_file(self):
        sys_ifname=self.__get_sys_ifname()
        interfaces_path=self.__get_cfg_fname()

        if not os.path.exists(interfaces_path):
            with open(interfaces_path, 'w') as f:
                pass
            self.update_interfaces_file()

        (fh, backup) = tempfile.mkstemp(dir='/tmp')
        os.close(fh)

        interfaces = debinterface.Interfaces(interfaces_path=interfaces_path, backup_path=backup)
        interfaces.removeAdapterByName(sys_ifname)

        ipv4_set = self.__get_ipv4_adapter_cfg()

        adapter = ipv4_set
        adapter.setAuto(self._enable)
        adapter.setHotplug(True)

        index = 0
        if ipv4_set:
            interfaces.addAdapter( ipv4_set.export(), index)
            index += 1

        interfaces.writeInterfaces()

        os.unlink(backup)

    def __update_mapping_file(self, pppox_idx, delete=0):
        sys_ifname = self.__get_sys_ifname()
        full_path = os.path.join(self.PPPOX_TO_HOST_IF_NAME_DIR_PATH, str(pppox_idx))
        if delete:
            if os.path.isfile(full_path):
                os.remove(full_path)
        else:
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            with open(full_path, "w") as f:
                f.write(sys_ifname)

    def __get_ipv4_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv4_set = debinterface.NetworkAdapter(sys_ifname)
        ipv4_set.setName(sys_ifname)
        ipv4_set.setAddrFam('inet')

        ipv4_set.setAddressSource('manual')
        # ipv4_set.setAddress(str('1.1.1.1'))
        # ipv4_set.setNetmask(str('255.255.255.255'))

        return ipv4_set

    def update_interfaces_file(self):
        with open(self.INTERFACES_FILE, 'a') as f:
            f.write('source %s\n' % self.__get_cfg_fname())

    @property
    def ip_pass(self):
        return self._ip_pass

    @ip_pass.setter
    def ip_pass(self, new):
        if new and not isinstance(new, bool):
            raise PPPoEClientIfErr('Bad IP-Passthrough type param', -1)
        self._ip_pass = new

    def _create_ippass_dhcpd_if(self):
        if not self._name:
            self.logger.error('Interface name is not set!')
            return

        self._create_ns()
        tap_id = int(self._name.replace('ppp',''))

        # macarr =  [0xfe, random.getrandbits(8),
        #    random.getrandbits(8), random.getrandbits(8),
        #    random.getrandbits(8), random.getrandbits(8)]
        macarr =  [0xfe, 0xfe, 0xfe, 0xfe, 0xfe, (tap_id + 150)]
        usedmac = ':'.join(["%02x" % x for x in macarr])
        nsname = '{}_dhcpd'.format(self._name)

        self.logger.debug('DHCPD NS name {}'.format(nsname))
        self.logger.debug('DHCPD ID {}'.format(tap_id))

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        (ret, tapidx)=vpp_if.create_tap_interface('ns{}' .format(self._name), id=tap_id+self.PPP_DHCPD_OFFSET, 
                                                            host_namespace_set=True,
                                                            host_namespace=nsname,
                                                            host_mac_addr=usedmac, 
                                                            mac_addr=usedmac,
                                                            host_mac_addr_set=True)

        if 0 != ret:
            self.logger.error('_create_ippass_dhcpd_if error - {}' .format(ret))
            return

        #Create bridge-domain
        vpp_if.add_to_bridge(self._tapidx, tapidx)
        vpp_if.ifup(tapidx)
        self._dhcpidx = tapidx
        self.logger.debug('IP-Pass DHCPD interface with id {} created in NS {}' .format(self._dhcpidx, nsname))
        del(vpp_if)
    
    def _delete_ippass_dhcpd_if(self):
        if not self._dhcpidx:
            self.logger.debug('IP-Pass dhcpd Interface does not exists!')
            return
        
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        vpp_if.ifdown(self._dhcpidx)
        vpp_if.delete_tap_interface(self._dhcpidx)

        self.logger.debug('IP-Pass DHCPD interface with id {} deleted' .format(self._dhcpidx))
        self._dhcpidx = None
        del(vpp_if)

    def _create_ns(self):
        nsname = '{}_dhcpd'.format(self._name)
        try:
            with NetNS(nsname) as netns:
                self.logger.debug('Netns already exists')
        except NetlinkError:
            NetNS(nsname, create=True)
            self.logger.debug('Netns created')

    def _delete_ns(self):
        nsname = '{}_dhcpd'.format(self._name)
        try:
            with NetNS(nsname) as netns:
                netns.remove()
                self.logger.debug('Netns removed')
        except NetlinkError:
            self.logger.debug('Netns does not exist')

    def enable_ippass(self):
        """ Called only in __init__, because ip-passthrough change requires reboot.
        """
        if self._ip_pass != True:
            self.logger.error('IP-Passthrough not enabled for interface {}' .format(self._name))
            return
        if self._tapidx == None:
            self.logger.error('PPPoE host interface do not exists')
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        vpp_if.enable_pppoe_ippass(host_idx=self._tapidx)

        # enable ARP term in BD / IP-Pass
        vpp_if.set_bd_arp_term(bd_id=self._tapidx)
        vpp_if.enable_ippass_arp_termination(host_idx=self._tapidx)

        self.logger.debug('Enable PPPoE IPPass on HOST {} with WAN {}'.format(self._tapidx,self._pppoe_idx))
        vpp_if.add_del_ippass_permit_host_entry(host_idx=self._tapidx, protocol='icmp')
        vpp_if.add_del_ippass_permit_host_entry(host_idx=self._tapidx, protocol='udp', dst_port=53)
        vpp_if.add_del_ippass_permit_host_entry(host_idx=self._tapidx, protocol='udp', dst_port=123)
        del(vpp_if)

    def disable_ippass(self):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        self.logger.debug('Disable PPPoE IPPass on HOST {} with WAN {}'.format(self._tapidx,self._pppoe_idx))
        vpp_if.add_del_ippass_permit_host_entry(host_idx=self._tapidx, protocol='icmp', add=0)
        vpp_if.add_del_ippass_permit_host_entry(host_idx=self._tapidx, protocol='udp', dst_port=53, add=0)
        vpp_if.add_del_ippass_permit_host_entry(host_idx=self._tapidx, protocol='udp', dst_port=123, add=0)

        vpp_if.disable_ippass_arp_termination(host_idx=self._tapidx)
        vpp_if.unset_bd_arp_term(bd_id=self._tapidx)

        del(vpp_if)

    def enable_ippass_wanif(self):
        if self._ip_pass != True:
            self.logger.error('IP-Passthrough not enabled for interface {}' .format(self._name))
            return
        if self._tapidx == None:
            self.logger.error('PPPoE host interface do not exists')
            return
        if self._pppoe_idx == None:
            self.logger.error('PPPoE interface do not exists')
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        vpp_if.enable_pppoe_ippass(host_idx=self._tapidx, wan_idx=self._pppoe_idx)
        del(vpp_if)
      
    def disable_ippass_wanif(self):
        if self._ip_pass != True:
            self.logger.error('IP-Passthrough not enabled for interface {}' .format(self._name))
            return
        if self._tapidx == None:
            self.logger.error('PPPoE host interface do not exists')
            return
        if self._pppoe_idx == None:
            self.logger.error('PPPoE interface do not exists')
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        vpp_if.disable_pppoe_ippass(host_idx=self._tapidx, wan_idx=self._pppoe_idx)
        del(vpp_if)

    def __get_ippass_mac(self):
        """ If ip-passthrough is enabled use previous used mac address from nova config
            and use it againg otherwise ucpe and VM will end up with different mac addresses
        """
        phy_ppp_offset = 100
        from configobj import ConfigObj

        macarr =  [0xfe, random.getrandbits(8),
           random.getrandbits(8), random.getrandbits(8),
           random.getrandbits(8), random.getrandbits(8)]
        
        if self._ip_pass != True:
            return ':'.join(["%02x" % x for x in macarr])
        
        physname = str("physnet%s" % (int(self._name.replace("ppp", "")) + phy_ppp_offset))
        if not os.path.isfile('/etc/neutron/plugins/ml2/vpp_agent.ini.agent'):
            return ':'.join(["%02x" % x for x in macarr])
        
        cfg_obj = ConfigObj("/etc/nova/nova.conf.agent")
        if 'physnet_mac_mapping' not in cfg_obj['neutron']:
            return ':'.join(["%02x" % x for x in macarr])
        physnets = cfg_obj['neutron']['physnet_mac_mapping']
        isphysnet = False
        for physnet in physnets:
            if physname + '-' in physnet:
                isphysnet = True
                break
        if isphysnet == True:
            return str(physnet).strip(physname + '-')
        
        return ':'.join(["%02x" % x for x in macarr])
