import binascii
import socket
import time
import vppapi_if.vppapi_if as vppif
import netifaces  as net

class SmartPPClassifier(object):
    
    ETH_TYPE_IPv4 = '080045'
    BGP_HOST_MASK = 'ffffff0000000000000000ff0000ffffffff00000000ffff0000000000000000'
    BGP_WAN_MASK  = 'ffffff0000000000000000ff000000000000ffffffff0000ffff000000000000'

    def __init__(self, wan_interface=None, ippass_position=0, logger=None):
        self._logger = logger
        
        self._wan_interface = wan_interface    
        self._protocol = None
        self._proto_wan_table = {}
        self._proto_host_table = {}

        self._tracehost_interface = None
        self._tracehost_probe_id = None

        self._enable = False

        self._skip_classify_tables = ippass_position

        self._protocol_handlers = {
            "BGP":  self._configure_bgp_smartpp,
            "OSPF": self._configure_ospf_startpp,
        }

        # Different from _enable 
        self._deactivated = False

    def _prerequisites(self):
        return [self._wan_interface, self._protocol, self._enable]

    # Checks on every config change wether to
    # activate the smart protocol passthrough
    def activator(setter_func):
        def wrapper(self, value):
            setter_func(self, value)
            self._activate()
        return wrapper

    @property
    def wan_interface(self):
        return self._wan_interface

    @wan_interface.setter
    @activator
    def wan_interface(self, value):
        self._wan_interface = value

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    @activator
    def protocol(self, value):
        self._protocol = value

        self._fetch_classify_table_ids()

    @property
    def tracehost_interface(self):
        return self._tracehost_interface

    @tracehost_interface.setter
    @activator
    def tracehost_interface(self, value):
        self._tracehost_interface = value

    @property
    def tracehost_probe_id(self):
        return self._tracehost_probe_id

    @tracehost_probe_id.setter
    @activator
    def tracehost_probe_id(self, value):
        self._tracehost_probe_id = int(value)

    @property
    def enable(self):
        return self._enable

    @enable.setter
    @activator
    def enable(self, value):
        # value comes as string in some cases
        self._enable = value == 'True' or value == True
        
        if not self._enable:
            self._deactivate()

    def _activate(self):
        if all(self._prerequisites()): 
            self._deactivate()
            self._configure_smartpp()

            self._deactivated = False

    def _deactivate(self):
        if not self._deactivated:
            self._configure_smartpp(deactivate=True)
            self._deactivated = True

    def _configure_smartpp(self, deactivate=False):
        self._protocol_handlers[self._protocol](deactivate)

    def _configure_ospf_startpp(self, deactivate=False):
        self._logger.info(f'{"Deactivating" if deactivate else "Activating"} OSPF SmartPPon interface {self.wan_interface}')


    def _configure_bgp_smartpp(self, deactivate=False):
        self._logger.info(f'{"Deactivating" if deactivate else "Activating"} BGP SmartPP on interface {self.wan_interface}')
        vpp_if = vppif.VppApiInterface('smartpp')

        wan_ip = self._get_wan_ip()
        if not wan_ip:
            self._logger.error(f'Could not get IP address from {self._wan_interface}.')
            del(vpp_if)
            return
        
        match_host = f'0800450000000000000000060000{self.ip_to_hex(wan_ip)}0000000000b30000000000000000'.rstrip("0")
        match_wan =  f'080045000000000000000006000000000000{self.ip_to_hex(wan_ip)}000000b3000000000000'.rstrip("0")

        vpp_if.add_del_classify_session(self._proto_host_table[self._protocol], match_host, is_add=not deactivate)
        vpp_if.add_del_classify_session(self._proto_wan_table[self._protocol], match_wan, is_add=not deactivate)
        del(vpp_if)

    def _fetch_classify_table_ids(self):
        vpp_if = vppif.VppApiInterface('smartpp')
        count = vpp_if.get_classify_tables_count()
        for i in range(count - 1, 1, -1): # Skip table 0
            table = vpp_if.get_classify_table(i)
            mask = table['mask'].hex()
            if SmartPPClassifier.BGP_HOST_MASK == mask:
                self._logger.info(f'BGP_HOST_MASK {table["table_id"]}')
                self._proto_host_table[self._protocol] = table['table_id']
                continue
            if SmartPPClassifier.BGP_WAN_MASK == mask:
                self._logger.info(f'BGP_WAN_MASK {table["table_id"]}')
                self._proto_wan_table[self._protocol] = table['table_id']
                
            if self._proto_wan_table and self._proto_host_table:
                if self._skip_classify_tables == 0:
                    del(vpp_if)
                    return
                
                self._proto_wan_table.clear()
                self._proto_host_table.clear()
                self._skip_classify_tables -= 1
        del(vpp_if)

    def _get_wan_ip(self):
        # TODO:  ipv6
        for retry in range(0, 5):
            try:
                return net.ifaddresses(f'br-{self._wan_interface}')[net.AF_INET][0]['addr']
            except Exception:
                retry += 1
                time.sleep(1/retry)
        
        return None

    @staticmethod
    def ip_to_hex(ip):
        return binascii.hexlify(socket.inet_aton(ip)).decode("utf-8")
