from sysrepo_interfaces import sysrepo_interfaces3 as srif
import json
import logging
from . import smartpp_classifier as spp_c

class SmartPPManager(object):

    def __init__(self,):
        self._init_logger()

        self._classifiers = []

        self._init_ippass_position_map()
        self._register_config_handlers()

    def _on_deleted_config(self, change):
        wan, protocol = self._fetch_wan_and_protocol_from_xpath(change['xpath'])
        self._delete_classifier(wan, protocol)

        self._debug_classifiers()

    def _on_modified_config(self, change):
        for classifier in self._classifiers:
                self._logger.info(f' classifier modified {classifier}')

    def _on_created_config(self, change):
        wan, protocol = self._fetch_wan_and_protocol_from_xpath(change['xpath'])
        
        classifier = self._get_classifier(wan, protocol)

        params = self.multiple_new_values(change['new_value'])
        if params:
            for param, value in params.items():
                self._apply_param(classifier, param, value)
        else:
            param = self._fetch_param_from_xpath(change['xpath'])
            self._apply_param(classifier, param, change['new_value'])

        self._debug_classifiers()

    def _register_config_handlers(self):
        self._config_operation_handlers = {
            "DELETED":  self._on_deleted_config,
            "MODIFIED": self._on_created_config, # Works the same as CREATED
            "CREATED":  self._on_created_config,
        }
    
    def call_config_handler_running(self, change):
        self._logger.info(f' SmartPP RUNNING {change}')
        xpath = next(iter(change.values()))
        operation = xpath['operation']

        if operation in self._config_operation_handlers:
            return self._config_operation_handlers[operation](xpath)
        
        return None
    
    def call_config_handler_startup(self, change):
        self._logger.info(f' SmartPP Startup {change}')
        for smartpp_entry in change:
            wan = smartpp_entry['wan-interface']
            protocol = smartpp_entry['protocol']

            classifier = self._get_classifier(wan, protocol)

            for param, value in smartpp_entry.items():
                if param == 'wan-interface':
                    continue
                self._apply_param(classifier, param, value)

        self._debug_classifiers()


    #   Gets the specified setter attribute from the classier
    #   and sets given value 
    def _apply_param(self, classifier_obj, param_name, value):
        setter_name = self._cli_param_to_setter(param_name)
        try:
            setattr(classifier_obj, setter_name, value)
        except AttributeError:
            pass

    def _get_classifier_by_iface_and_proto(self, iface, protocol):
        for classifier in self._classifiers:
            # Properties
            if classifier.wan_interface == iface and \
                    classifier.protocol == protocol:
                return classifier
        
        return None
    
    # Returns any associated smartpp instances with this probe
    def _get_classifiers_by_probe(self, wan, probe_id):
        classifiers = []
        for classifier in self._classifiers:
            if classifier.tracehost_interface == wan and \
                classifier.tracehost_probe_id == probe_id:
                classifiers.append(classifier)

        return classifiers
    
    # Returns existing classifier 
    # or creates and registers new one 
    def _get_classifier(self, wan, protocol):
        classifier = self._get_classifier_by_iface_and_proto(wan, protocol)
        if not classifier:
            position = self._get_ippass_position(wan)
            classifier = spp_c.SmartPPClassifier(wan, position, self._logger)
            self._register_classifier(classifier)
        
        return classifier

    def _delete_classifier(self, wan, protocol):
        for idx, classifier in enumerate(list(self._classifiers)): # make a copy of the list
             if classifier.wan_interface == wan and \
                classifier.protocol == protocol:
                  classifier._deactivate()
                  self._classifiers.pop(idx)

    # Changes '-' with '_' to make
    # classifier setters to match cli command params
    # tracehost-interface , tracehost-probe-id etc.
    def _cli_param_to_setter(self, param):
        return param.replace('-', '_')
    
    def _register_classifier(self, classifier):
        self._classifiers.append(classifier)
    
    def _fetch_wan_and_protocol_from_xpath(self, xpath):
        name = "name='"
        proto = "protocol='"
        end = "']"

        xpath = xpath.split('smart-protocol-passthrough')
        s1 = xpath[0]
        s2 = xpath[1]

        wan = s1[s1.find(name)+len(name):s1.rfind(end)]
        protocol = s2[s2.find(proto)+len(proto):s2.rfind(end)]

        return (wan, protocol)

    def _fetch_param_from_xpath(self, xpath):
        return xpath.split('/')[-1]
    
    @staticmethod
    def multiple_new_values(change):
        try:
            change = change.replace("'", "\"")
            change = change.replace("True", "true")
            change = change.replace("False", "false")
            values = json.loads(change)
        except ValueError:
            return False
        if not isinstance(values, dict):
             return False
        return values
    
    # This is needed when classifiers fetch classify tables 
    # as for every ip-passthrough interface there is a new set
    # of classify tables . So classifiers will skip to their position
    def _init_ippass_position_map(self):
        sri = srif.SysrepoInterfaces(app_name='smartpp')
        
        self._ippass_position_map = {}
        position = 0

        interfaces = sri.get_interfaces(startup=True)
        for interface, config in interfaces.items():
            if config.get('ip-passthrough', False):
                self._ippass_position_map[interface] = position
                position += 1
        
        interfaces = self._ippass_position_map.keys()
        positions = reversed(self._ippass_position_map.values())

        self._ippass_position_map = {k:v for k,v in zip(interfaces, positions)}

    def _get_ippass_position(self, wan):
        return self._ippass_position_map[wan]

    def handle_th_probe_result(self, wan, probe, succeeded, needs_change):
        classifiers = self._get_classifiers_by_probe(wan, probe['id'])
        if not classifiers:
            return 
        
        for classifier in classifiers:
            if succeeded == needs_change:
                classifier._activate()
            else:
                classifier._deactivate()

    def _debug_classifiers(self):
         for classifier in self._classifiers:
                self._logger.info(f' classifier config {classifier._wan_interface} \
                                  {classifier._protocol} {classifier._tracehost_interface} \
                                  {classifier._tracehost_probe_id} {classifier._enable}')

    def _init_logger(self):
        self._logger = logging.getLogger(name='smartpp')
        self._logger.setLevel(level=logging.INFO)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self._logger.addHandler(syslog_hnd)
        self._logger.propagate = False

    