import unittest
import struct
import subprocess
import sys
import socket
import time
#from _thread import start_new_thread
import threading
import os
import math
import fcntl


def TunTap(nic_type,nic_name=None, logger=None):
    '''
    Use TunTap to create tuntap loopback in OVS
    
    input:
        nic_type:  must be "Tun" or "Tap"
        nic_name:  device name, default is None,
                    if None will auto generate, can be obtained by tap.name
                    else will reuse the name of given
    
    after tap create, can be config(ip,mask), then can be tested with read or write
    '''
    tap = Tap(nic_type,nic_name, logger=logger)
    tap.create()
    return tap

class Tap(object):
    def __init__(self,nic_type,nic_name=None, logger=None):
        self.nic_type = nic_type
        self.name = nic_name
        self.mac = b"\x00"*6
        self.handle = None
        self.ip4 = None
        self.ip6 = None
        self.read_lock = threading.Lock()
        self.write_lock = threading.Lock()
        self.quitting = False
        self.logging = logger

    def create(self):
        TUNSETIFF = 0x400454ca
        TUNSETOWNER = 0x400454cc
        TUNSETGROUP = 0x400454ce
        TUNSETPERSIST = 0x400454cb
        IFF_TUN = 0x0001
        IFF_TAP = 0x0002
        IFF_MULTI_QUEUE = 0x0100
        IFF_NO_PI = 0x1000
        O_RDWR = 0x2
        # Open TUN device file
        tun = os.open('/dev/net/tun', O_RDWR)
        if not tun:
            return None
        if self.nic_type == "Tap":
            flags = IFF_TAP | IFF_NO_PI
        if self.nic_type == "Tun":
            flags = IFF_TUN | IFF_NO_PI
        if self.name:
            ifr_name = self.name.encode() + b'\x00'*(16-len(self.name.encode()))
        else:
            ifr_name = b'\x00'*16
        ifr = struct.pack('16sH22s', ifr_name , flags,b'\x00'*22)
        try:
            ret = fcntl.ioctl(tun, TUNSETIFF, ifr)
            self.logging.debug("%s %s"%(ifr,ret))
            dev, _ = struct.unpack('16sH', ret[:18])
            dev = dev.decode().strip("\x00")
            self.name = dev
            fcntl.ioctl(tun, TUNSETOWNER, struct.pack("H",1000))
            fcntl.ioctl(tun, TUNSETGROUP, struct.pack("H",1000))
            fcntl.ioctl(tun, TUNSETPERSIST, struct.pack("B",True))
            self.handle = tun
        except OSError:
            self.logging.error("Device or resource busy")

        if self.handle:
            return self
        else:
            return None


    def config(self, ip4=None, ip6=None):
        '''
        config device's ip and mask

        input:
            ip4:  ipaddress string, such as "192.168.1.5/24"
            ip6:  ipaddress string, such as "ff11::3/128"

        return :
            None  if failure
            self  if success

        after tap configed,then canbe read or write ,please refer
        '''
        self.logging.debug(f"tuntap_ovs: IFUP loopback interface {self.name}")
        subprocess.check_call('ip link set '+self.name+' up', shell=True)
        try:
            if ip4 is not None:
                self.logging.debug("tuntap_ovs: IPv4 are given")
                self.ip4 = ip4
                self.logging.debug(f"tuntap_ovs: Set IPv4 for interface {self.name} : {ip4}")
                subprocess.check_call('ip addr add ' + self.ip4 + " dev "+ self.name , shell=True)
            if ip6 is not None:
                self.logging.debug("tuntap_ovs: IPv6 are given")
                self.ip6 = ip6
                self.logging.debug(f"tuntap_ovs: Set IPv6 for interface {self.name} : {ip6}")
                subprocess.check_call('ip -6 addr add ' + self.ip6 + " dev "+ self.name , shell=True)
        except:
            self.logging.error(f"Error on tap ovs config : name:{self.name}, ip4:{self.ip4}, ip6:{self.ip6}")
            self.close()
            return None
        return  self

    def close(self):
        self.quitting = False
        if self.handle:
            try: 
                os.close(self.handle)
            except OSError:
                pass
        try:
            if self.ip4 is not None:
                subprocess.check_call('ip addr delete ' + self.ip4 + " dev "+ self.name , shell=True)
                self.ip4 = None
            if self.ip6 is not None:
                subprocess.check_call('ip -6 addr delete ' + self.ip6 + " dev "+ self.name , shell=True)
                self.ip6 = None
            if self.ip4 is None and self.ip6 is None:
                subprocess.check_call('ip tuntap delete mode tap ' + self.name , shell=True)

        except Exception as e:
            self.logging.debug(e)
            pass
        pass
    
    # READ / WRITE tests if needed
    """
    def read(self,size=1522):
        '''
        read device data with given size

        input:
            size:  read max size , int . such as size = 1500

        return :
            bytes:

        '''
        self.read_lock.acquire()
        data = os.read(self.handle,size)
        self.read_lock.release()
        return data
        pass

    def write(self,data):
        '''
        write data to device

        input:
            data:  byte[] . such as data = b'\x00'*100

        return :
            int:  writed bytes

        '''
        result = 0
        self.write_lock.acquire()
        try:
            result = os.write(self.handle,data)
        except:
            pass
        self.write_lock.release()
        return result
    """
