#!/usr/bin/env python3
import os
import logging
import logging.handlers
import vppapi_if.vppapi_if as vppif
import ipaddress
import debinterface
import tempfile
from enum import IntEnum

LOG_NAME = 'vxlan-cls'

class BOOT_PROTO(IntEnum):
    NONE=1
    DHCP=2
    DHCP6=3
    DHCP46=4

class TUNNEL_TYPE(IntEnum):
    VXLAN=1
    VXLAN_GPE=2

class VxLanIfErr(Exception):
    def __init__(self, message, code):
        super(VxLanIfErr, self).__init__(message)
        self.code = code

class VxLanInterface(object):
    BR_PREFFIX='br-'    
    VXLAN_OFFSET=500
    LOOP_OFFSET=600
    CFG_FILE_PATH='/etc/network/interfaces.d/ifcfg-{}'
    INTERFACES_FILE='/etc/network/interfaces'

    def __init__(self, name, vni=None, lvtep=None, lvtepv6=None, rvteps=[], tun_type=TUNNEL_TYPE.VXLAN, logger=None, log_level=logging.ERROR, 
                    shg=100, mcast_if=None, mcast_ip=None ):
        self._management = False
        self._enable = False
        self._ipv4 = None
        self._ipv6 = None
        self._bootproto = BOOT_PROTO.NONE

        self.logger = None
        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level ) 

        self._name = name
        self.logger.debug('VxLanInterface for \'{}\' - init vni={}, lvtep={}, rvteps={}'.format(self._name, vni, lvtep, rvteps))

        if vni and not isinstance(vni, int):
            raise VxLanIfErr('Bad type of VNI, expect int', -1)
        
        self._vni = vni

        if shg and not isinstance(shg, int):
            raise VxLanIfErr('Bad type of SHG, expect int', -1)
        
        self._shg = shg

        if lvtep and not isinstance(lvtep, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            raise VxLanIfErr('Bad type of lvtep, expect IPv4Address or IPv6Address', -1)
        
        self._lvtep = lvtep

        if lvtepv6 and not isinstance(lvtepv6, (ipaddress.IPv6Address)):
            raise VxLanIfErr('Bad type of lvtepv6, expect IPv6Address', -1)

        self._lvtepv6 = lvtepv6

        if rvteps and not isinstance(rvteps, list):
            raise VxLanIfErr('Bad param type of rvteps , expect list of IPv4Address or Ipv6Address', -1)

        rv4 = []
        rv6 = []
        for rvtep in rvteps:
            if isinstance(rvtep, ipaddress.IPv4Address):
                rv4.append(rvtep)
            elif isinstance(rvtep, ipaddress.IPv6Address):
                rv6.append(rvtep)
            else:
                raise VxLanIfErr('Bad param type of rvteps element, expect IPv4Address or Ipv6Address', -1)

        self._rvtep = set()
        if rv4:
            self._rvtep.update(rv4)

        self._rvtepv6 = set()
        if rv6:
            self._rvtepv6.update(rv6)

        self._vpp_bridx = None
        self._vpp_vxlans = {}

        if tun_type and not isinstance(tun_type, TUNNEL_TYPE):
            raise VxLanIfErr('Bad tunnel type', -1)
        
        self._tun_type = tun_type

        self._create_vxlan_bd()
        self.logger.debug( "VNI {}, LVTEP {} / {} num of RVTEP {} / {}".format(self._vni, self._lvtep, self._lvtepv6, len(self._rvtep), len(self._rvtepv6)) )

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        if self._vni and self._lvtep and len(self._rvtep):
            for rvtep in self._rvtep:
                tun_idx = self._create_vxlan_if(rvtep, self._lvtep, vpp_if=vpp_if)
                if tun_idx:
                    self._add_ifid_to_bd(tun_idx, vpp_if=vpp_if, add_shg=True)
                    self._vpp_vxlans[rvtep] = tun_idx

        if self._vni and self._lvtepv6 and len(self._rvtepv6):
            for rvtep in self._rvtepv6:
                tun_idx = self._create_vxlan_if(rvtep, self._lvtepv6, vpp_if=vpp_if)
                if tun_idx:
                    self._add_ifid_to_bd(tun_idx, vpp_if=vpp_if, add_shg=True)
                    self._vpp_vxlans[rvtep] = tun_idx

        if mcast_if == None:
            self._mcast_if = None
        else:
            self._mcast_if = self.find_mcast_if_vpp(hostifname=mcast_if,vpp_if=vpp_if)

        if mcast_ip and not isinstance(mcast_ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            raise VxLanIfErr('Bad param type expect ipaddress.IPv4Address, ipaddress.IPv6Address or None', -1)
        if mcast_ip and not ( str(mcast_ip).startswith('239') or str(mcast_ip).startswith('ff')):
            raise VxLanIfErr('Bad IP address expect multicast in range 239.0.0.0/8 for IPv4 and FF00::/8 for IPv6', -1)
        self._mcast_ip = mcast_ip

        if self._vni and self._mcast_if and self._mcast_ip and (self._lvtep or self._lvtepv6):
            self.logger.debug( "VNI {}, LVTEP {} / {} MCAST IF {} MCAST ADDR --".format(self._vni, self._lvtep, self._lvtepv6, self._mcast_if) )
            
            mcast_idx = self._create_vxlan_mcast_if(vpp_if=vpp_if)
            if mcast_idx and mcast_idx != 0xffffffff:
                self._add_ifid_to_bd(mcast_idx, vpp_if=vpp_if, add_shg=True)
                self._vpp_vxlans[self._mcast_ip] = mcast_idx

        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if)

        self.__update_cfg_file()
        self.down_sys_if()
        del(vpp_if)


    def __del__(self):
        self.logger.debug("VxLanInterface {} __del__".format(self._name))

        interfaces = debinterface.Interfaces(interfaces_path=self.__get_cfg_fname())
        interfaces.downAdapter(self.__get_sys_ifname())

        os.unlink(self.__get_cfg_fname())

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if, delete=True)

        for vxlan_idx in self._vpp_vxlans.values():
            if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                vpp_if.vxlan_gpe_delete_by_idx(vxlan_idx)
            else:
                vpp_if.vxlan_delete_by_idx(vxlan_idx)


        if self._vpp_bridx:
            vpp_if.delete_tap_interface(self._vpp_bridx)
            vpp_if.delete_bridge_domain(self._vpp_bridx)



    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)

    def __get_sys_ifname(self):
        return '{}{}'.format(self.BR_PREFFIX, self._name)

    def __get_cfg_fname(self):
        return self.CFG_FILE_PATH.format(self.__get_sys_ifname())

    def __get_ipv4_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv4_set = debinterface.NetworkAdapter(sys_ifname)
        ipv4_set.setName(sys_ifname)
        ipv4_set.setAddrFam('inet')

        if self._bootproto == BOOT_PROTO.NONE:
            if self._ipv4:
                ipv4_set.setAddressSource('static')
                ipv4_set.setAddress(str(self._ipv4.ip))
                ipv4_set.setNetmask(str(self._ipv4.network.netmask))
            else:
                ipv4_set.setAddressSource('manual')
        elif self._bootproto == BOOT_PROTO.DHCP or self._bootproto == BOOT_PROTO.DHCP46:
            ipv4_set.setAddressSource('dhcp')
        else:
            return None

        return ipv4_set
        
    def __get_ipv6_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv6_set = debinterface.NetworkAdapter(sys_ifname)
        ipv6_set.setName(sys_ifname)
        ipv6_set.setAddrFam('inet6')

        if self._bootproto == BOOT_PROTO.NONE:
            if self._ipv6:
                ipv6_set.setAddressSource('static')
                ipv6_set.setAddress(str(self._ipv6.ip))
                ipv6_set.setNetmask(str(self._ipv6.network.prefixlen))
            else:
                return None
        elif self._bootproto == BOOT_PROTO.DHCP6 or self._bootproto == BOOT_PROTO.DHCP46:
            ipv6_set.setAddressSource('dhcp')
        else:
            return None

        ipv6_set.appendPreUp('echo 0 > /proc/sys/net/ipv6/conf/{}/accept_dad'.format(sys_ifname))
        return ipv6_set

    def __update_cfg_file(self):
        sys_ifname=self.__get_sys_ifname()
        interfaces_path=self.__get_cfg_fname()

        if not os.path.exists(interfaces_path):
            with open(interfaces_path, 'w') as f:
                pass
            self.update_interfaces_file()

        (fh, backup) = tempfile.mkstemp(dir='/tmp')
        os.close(fh)

        interfaces = debinterface.Interfaces(interfaces_path=interfaces_path, backup_path=backup)
        interfaces.removeAdapterByName(sys_ifname)

        ipv4_set = self.__get_ipv4_adapter_cfg()
        ipv6_set = self.__get_ipv6_adapter_cfg()

        adapter = ipv4_set if ipv4_set else ipv6_set

        adapter.setAuto(self._enable)
        adapter.setHotplug(True)

        index = 0
        if ipv4_set:
            interfaces.addAdapter( ipv4_set.export(), index)
            index += 1
        
        if ipv6_set:
            interfaces.addAdapter( ipv6_set.export(), index)

        interfaces.writeInterfaces()

        os.unlink(backup)

    def update_interfaces_file(self):
        with open(self.INTERFACES_FILE, 'a') as f:
            f.write('source %s\n' % self.__get_cfg_fname())

    @property
    def name(self):
        return self._name

    @property
    def vni(self):
        return self._vni

    @vni.setter
    def vni(self, new):
        self.logger.debug("Set VNI for {} to {}".format(self._name, new))
        if not isinstance(new, int) and new:
            raise VxLanIfErr('Expect integer or None', -1)
        
        if self._vni == new:
            #no change
            return

        vpp_if = vppif.VppApiInterface(LOG_NAME)

        self.__delete_all_vxlan_tunnels(vpp_if)
        self._vni = new
        self.__create_all_vxlan_tunnels(vpp_if)

    @property
    def lvtep(self):
        return self._lvtep

    @lvtep.setter
    def lvtep(self, new):
        if new != None:
            if new and not isinstance(new, ipaddress.IPv4Address):
                raise VxLanIfErr('Bad param type expect ipaddress.IPv4Address or None', -1)

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        self.__delete_all_ipv4_vxlan_tunnels(vpp_if)
        self._delete_vxlan_mcast_if(vpp_if)
        self._lvtep = new
        self.__create_all_ipv4_vxlan_tunnels(vpp_if=vpp_if)
        self.__create_mcast_tunnel(vpp_if=vpp_if)
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if)

    @property
    def lvtepv6(self):
        return self._lvtepv6

    @lvtepv6.setter
    def lvtepv6(self, new):
        if new and not isinstance(new, ipaddress.IPv6Address):
            raise VxLanIfErr('Bad param type expect ipaddress.IPv6Address or None', -1)

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        self.__delete_all_ipv6_vxlan_tunnels(vpp_if)
        self._delete_vxlan_mcast_if(vpp_if)
        self._lvtepv6 = new
        self.__create_all_ipv6_vxlan_tunnels(vpp_if=vpp_if)
        self.__create_mcast_tunnel(vpp_if=vpp_if)
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if)

    def add_rvteps(self, rvteps):
        rv4 = []
        rv6 = []
        if not isinstance(rvteps, list):
            raise VxLanIfErr("Parameter rvteps is not list", -1)
        
        for rvtep in rvteps:
            if isinstance(rvtep, ipaddress.IPv4Address):
                rv4.append(rvtep)
            elif isinstance(rvtep, ipaddress.IPv6Address):
                rv6.append(rvtep)
            else:
                raise VxLanIfErr("Element in rvteps is not ipaddress.IPv4Address or ipaddress.IPv6Address", -1 )
        
        vpp_if = vppif.VppApiInterface(LOG_NAME)
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if, delete=True)

        if len(self._rvtep.symmetric_difference(rv4)):
            if self._vni and self._lvtep and self._vpp_bridx:
                for rvtep in self._rvtep.symmetric_difference(rv4):
                    vxlan_idx = self._create_vxlan_if(rvtep, self._lvtep, vpp_if=vpp_if)
                    self._add_ifid_to_bd(vxlan_idx, vpp_if=vpp_if, add_shg=True)
                    if vxlan_idx != 0xffffffff:
                        self._vpp_vxlans[rvtep] = vxlan_idx
            self._rvtep.symmetric_difference_update(rv4)

        if len(self._rvtepv6.symmetric_difference(rv6)):
            if self._vni and self._lvtepv6 and self._vpp_bridx:
                for rvtep in self._rvtepv6.symmetric_difference(rv6):
                    vxlan_idx = self._create_vxlan_if(rvtep, self._lvtepv6, vpp_if=vpp_if)
                    self._add_ifid_to_bd(vxlan_idx, vpp_if=vpp_if, add_shg=True)
                    if vxlan_idx != 0xffffffff:
                        self._vpp_vxlans[rvtep] = vxlan_idx
            self._rvtepv6.symmetric_difference_update(rv6)

        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if)

        self.logger.debug("After add RVTEPS {} and {}".format(self._rvtep, self._rvtepv6))

    def del_rvteps(self, rvteps):
        rv4 = []
        rv6 = []
        if not isinstance(rvteps, list):
            raise VxLanIfErr("Parameter rvteps is not list", -1)
        
        for rvtep in rvteps:
            if isinstance(rvtep, ipaddress.IPv4Address):
                rv4.append(rvtep)
            elif isinstance(rvtep, ipaddress.IPv6Address):
                rv6.append(rvtep)
            else:
                raise VxLanIfErr("Element in rvteps is not ipaddress.IPv4Address or ipaddress.IPv6Address", -1 )

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if, delete=True)

        if len(self._rvtep.intersection(rv4)):
            if self._vni and self._lvtep and self._vpp_bridx:
                for rvtep in self._rvtep.intersection(rv4):
                    vxlan_idx = self._vpp_vxlans.get( rvtep, None )
                    if not vxlan_idx:
                        continue

                    vpp_if.del_from_bridge(self._vpp_bridx, vxlan_idx)
                    vpp_if.vxlan_delete_by_idx(vxlan_idx)
                    self._vpp_vxlans.pop(rvtep)

            self._rvtep.difference_update(rv4)

        if len(self._rvtepv6.intersection(rv6)):
            if self._vni and self._lvtepv6 and self._vpp_bridx:
                for rvtep in self._rvtepv6.intersection(rv6):
                    vxlan_idx = self._vpp_vxlans.get( rvtep, None )
                    if not vxlan_idx:
                        continue

                    vpp_if.del_from_bridge(self._vpp_bridx, vxlan_idx)
                    vpp_if.vxlan_delete_by_idx(vxlan_idx)
                    self._vpp_vxlans.pop(rvtep)

            self._rvtepv6.difference_update(rv6)

        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if)

        self.logger.debug("After del RVTEPS {} and {}".format(self._rvtep, self._rvtepv6))

    def clean_rvteps(self):
        self.logger.debug("Clean all rvteps")
        self.__delete_all_vxlan_tunnels()
        self._rvtep.clear()

    def _create_vxlan_bd(self):
        if not self._name:
            self.logger.error('Interface name is not set!')
            return

        vpp_if = vppif.VppApiInterface(LOG_NAME)

        #get tap ind
        tap_id = int(self._name.replace('vxlan',''))
        self.logger.debug('vxlan ID {}'.format(tap_id))

        (ret, bridx)=vpp_if.create_tap_interface('{}{}'.format(self.BR_PREFFIX,self._name), id=tap_id+self.VXLAN_OFFSET, host_mac_addr=None)

        if 0 == ret:
            vpp_if.ifup(bridx)

        ret = vpp_if.create_bridge_domain(bridx)
        vpp_if.add_to_bridge(bridx,bridx)
        self._vpp_bridx = bridx

    def _add_ifid_to_bd(self, ifidx, vpp_if=None, add_shg=False):
        if not self._vpp_bridx:
            return -1

        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)

        ret = vpp_if.add_to_bridge(self._vpp_bridx,ifidx, shg=self._shg)
        return ret

    def _create_vxlan_if(self, rvtep, lvtep, vpp_if=None):
        if not self._name or not self._vni or not lvtep:
            self.logger.error("Configuration is not full")
            return None

        if not self._vpp_bridx:
            self.logger.error("VPP bridge is not configured")
            return None

        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)

        if type(rvtep) != type(lvtep):
            return None

        if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
            (ret, sw_if_idx) = vpp_if.vxlan_gpe_create( self.vni, lvtep, rvtep)
        else:
            (ret, sw_if_idx) = vpp_if.vxlan_create( self.vni, lvtep, rvtep)

        return sw_if_idx

    def _create_vxlan_mcast_if(self, vpp_if=None):
        local_vtep = None
        local_vtep = self._lvtepv6 if isinstance(self._mcast_ip, ipaddress.IPv6Address) else self._lvtep

        if not self._name or not self._vni or not local_vtep or not self._mcast_if or not self._mcast_ip:
            self.logger.error("MC Configuration is not full")
            return None

        if not self._vpp_bridx:
            self.logger.error("VPP bridge is not configured")
            return None

        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)

        if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
            (ret, sw_if_idx) = vpp_if.vxlan_gpe_create( vni=self.vni, lvtep=local_vtep, rvtep=self._mcast_ip, mcast_if_index=self._mcast_if)
        else:
            (ret, sw_if_idx) = vpp_if.vxlan_create( vni=self.vni, lvtep=local_vtep, rvtep=str(self._mcast_ip), mcast_if_index=self._mcast_if)

        self.logger.debug("_create_vxlan_mcast_if ret - {} - idx - {}" .format(ret, sw_if_idx))
        return sw_if_idx

    def _delete_vxlan_mcast_if(self, vpp_if=None):
        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)
        
        vxlan_idx = self._vpp_vxlans.get( self._mcast_ip, None )
        if not vxlan_idx:
            return

        vpp_if.del_from_bridge(self._vpp_bridx, vxlan_idx)
        if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
            vpp_if.vxlan_gpe_delete_by_idx(vxlan_idx)
        else:
            vpp_if.vxlan_delete_by_idx(vxlan_idx)
        
        self._vpp_vxlans.pop(self._mcast_ip)

    def __delete_all_vxlan_tunnels(self, vpp_if=None):
        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)

        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if, delete=True)

        for vxlan_idx in self._vpp_vxlans.values():
            vpp_if.del_from_bridge(self._vpp_bridx, vxlan_idx)
            if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                vpp_if.vxlan_gpe_delete_by_idx(vxlan_idx)
            else:
                vpp_if.vxlan_delete_by_idx(vxlan_idx)

        self._vpp_vxlans.clear()

    def __delete_all_ipv4_vxlan_tunnels(self, vpp_if=None):
        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)
        
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if, delete=True)

        if len(self._rvtep):
            if self._vni and self._vpp_bridx:
                for rvtep in self._rvtep:
                    vxlan_idx = self._vpp_vxlans.get( rvtep, None )
                    if not vxlan_idx:
                        continue

                    vpp_if.del_from_bridge(self._vpp_bridx, vxlan_idx)
                    if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                        vpp_if.vxlan_gpe_delete_by_idx(vxlan_idx)
                    else:
                        vpp_if.vxlan_delete_by_idx(vxlan_idx)
                    self._vpp_vxlans.pop(rvtep)
        self.logger.debug('__delete_all_ipv4_vxlan_tunnels - AFTER {}' .format(self._vpp_vxlans.values()))

    def __delete_all_ipv6_vxlan_tunnels(self, vpp_if=None):
        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)
        
        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if, delete=True)

        if len(self._rvtepv6):
            if self._vni and self._vpp_bridx:
                for rvtep in self._rvtepv6:
                    vxlan_idx = self._vpp_vxlans.get( rvtep, None )
                    if not vxlan_idx:
                        continue

                    vpp_if.del_from_bridge(self._vpp_bridx, vxlan_idx)
                    if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                        vpp_if.vxlan_gpe_delete_by_idx(vxlan_idx)
                    else:
                        vpp_if.vxlan_delete_by_idx(vxlan_idx)
                    self._vpp_vxlans.pop(rvtep)
        self.logger.debug('__delete_all_ipv6_vxlan_tunnels - AFTER {}' .format(self._vpp_vxlans.values()))

    def __create_mcast_tunnel(self, vpp_if=None):
        if not self._mcast_ip or not self.mcast_if:
            return
        
        mcast_idx = self._create_vxlan_mcast_if(vpp_if=vpp_if)
        if mcast_idx and mcast_idx != 0xffffffff:
            self._add_ifid_to_bd(mcast_idx, vpp_if=vpp_if, add_shg=True)
            self._vpp_vxlans[self._mcast_ip] = mcast_idx

        self.logger.debug('__create_mcast_tunnel - AFTER {}' .format(self._vpp_vxlans.values()))

    def __create_all_ipv4_vxlan_tunnels(self, vpp_if=None):
        if not self._vni or not self._vpp_bridx or not self._rvtep:
            return
        
        for rvtep in self._rvtep:
            vxlan_idx = self._create_vxlan_if(rvtep, self._lvtep, vpp_if=vpp_if)
            self._add_ifid_to_bd( vxlan_idx, vpp_if=vpp_if, add_shg=True)
            self._vpp_vxlans[rvtep] = vxlan_idx
        
        self.logger.debug('__create_all_ipv4_vxlan_tunnels - AFTER {}' .format(self._vpp_vxlans.values()))

    def __create_all_ipv6_vxlan_tunnels(self, vpp_if=None):
        if not self._vni or not self._vpp_bridx or not self._rvtepv6:
            return
        
        for rvtep in self._rvtepv6:
            vxlan_idx = self._create_vxlan_if(rvtep, self._lvtepv6, vpp_if=vpp_if)
            self._add_ifid_to_bd( vxlan_idx, vpp_if=vpp_if, add_shg=True)
            self._vpp_vxlans[rvtep] = vxlan_idx
        
        self.logger.debug('__create_all_ipv6_vxlan_tunnels - AFTER {}' .format(self._vpp_vxlans.values()))

    def __create_all_vxlan_tunnels(self, vpp_if=None):
        if not self._vni or not self._vpp_bridx:
            return

        if not vpp_if:
            vpp_if = vppif.VppApiInterface(LOG_NAME)

        self.__create_all_ipv4_vxlan_tunnels(vpp_if=vpp_if)
        self.__create_all_ipv6_vxlan_tunnels(vpp_if=vpp_if)
        self.__create_mcast_tunnel(vpp_if=vpp_if)

        if self._vpp_vxlans:
            self.__rt_ctl_update(vpp_if=vpp_if)

    def __rt_ctl_update(self, vpp_if=None, delete=False):
        if 'yes' == os.getenv('RTCTL', ''):
            if self._lvtep:
                if not vpp_if:
                    vpp_if = vppif.VppApiInterface(LOG_NAME)

                remove = 1 if delete else 0
                if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                    ret = vpp_if.routectl_permit_proto( "udp", str(self._lvtep), dst_port=int(4790), disable=remove)
                else:
                    ret = vpp_if.routectl_permit_proto( "udp", str(self._lvtep), dst_port=int(4789), disable=remove)
                self.logger.debug("routectl_permit_proto -> ret {}".format(ret))

            if self._lvtepv6:
                if not vpp_if:
                    vpp_if = vppif.VppApiInterface(LOG_NAME)

                remove = 1 if delete else 0
                if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                    ret = vpp_if.routectl_permit_proto( "udp", str(self._lvtepv6), dst_port=int(4790), disable=remove)
                else:
                    ret = vpp_if.routectl_permit_proto( "udp", str(self._lvtepv6), dst_port=int(4789), disable=remove)
                self.logger.debug("routectl_permit_proto -> ret {}".format(ret))

            if self._mcast_ip:
                if not vpp_if:
                    vpp_if = vppif.VppApiInterface(LOG_NAME)

                remove = 1 if delete else 0
                if self._tun_type == TUNNEL_TYPE.VXLAN_GPE:
                    ret = vpp_if.routectl_permit_proto( "udp", str(self._mcast_ip), dst_port=int(4790), disable=remove)
                else:
                    ret = vpp_if.routectl_permit_proto( "udp", str(self._mcast_ip), dst_port=int(4789), disable=remove)
                self.logger.debug("routectl_permit_proto -> ret {}".format(ret))

    @property
    def management(self):
        return self._management

    @management.setter
    def management(self, new):
        if not isinstance(new, bool):
            raise VxLanIfErr("management expect boolean", -1)

        self._management = new

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, new):
        if not isinstance(new, bool):
            raise VxLanIfErr("enable expect boolean", -1)

        self.logger.debug("Set enable from {} to {} for {}".format(self._enable, new, self._name))
        if self._enable == new:
            return

        if self._enable:
            self.down_sys_if()

        self._enable = new
        self.__update_cfg_file()

        if new:
            self.up_sys_if()

    @property
    def bootproto(self):
        return self._bootproto

    @bootproto.setter
    def bootproto(self, new):
        if not isinstance(new, BOOT_PROTO ):
            raise VxLanIfErr("bootproto expect type BOOT_PROTO", -1)

        if self._bootproto == new:
            return

        self._bootproto = new

        if self._enable:
            self.down_sys_if()

        self.__update_cfg_file()

        if self._enable:
            self.up_sys_if()

    @property
    def ip4(self):
        return self._ipv4

    @ip4.setter
    def ip4(self, new):
        if new and not isinstance(new, ipaddress.IPv4Interface):
            raise VxLanIfErr("ipv4 expect ipaddress.IPv4Interface or None", -1)

        if self._ipv4 == new:
            return

        if self._bootproto == BOOT_PROTO.NONE and self._enable:
            self.down_sys_if()

        self._ipv4 = new
        self.__update_cfg_file()

        if self._bootproto == BOOT_PROTO.NONE and self._enable:
            self.up_sys_if()

    @property
    def ip6(self):
        return self._ipv6

    @ip6.setter
    def ip6(self, new):
        if new and not isinstance(new, ipaddress.IPv6Interface):
            raise VxLanIfErr("ipv4 expect ipaddress.IPv6Interface or None", -1)

        if self._ipv6 == new:
            return

        if self._bootproto == BOOT_PROTO.NONE and self._enable:
            self.down_sys_if()

        self._ipv6 = new
        self.__update_cfg_file()

        if self._bootproto == BOOT_PROTO.NONE and self._enable:
            self.up_sys_if()

    @property
    def tun_type(self):
        return self._tun_type

    @tun_type.setter
    def tun_type(self, new):
        if new  and not isinstance(new, TUNNEL_TYPE):
            raise VxLanIfErr('Bad tunnel param type', -1)

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        self.__delete_all_vxlan_tunnels(vpp_if)
        self._tun_type = new
        self.__create_all_vxlan_tunnels(vpp_if)

    @property
    def shg(self):
        return self._shg

    @shg.setter
    def shg(self, new):
        self.logger.debug("Set SHG for {} to {}".format(self._name, new))
        if not isinstance(new, int) and new:
            raise VxLanIfErr('Expect integer', -1)
        
        if self._shg == new:
            return

        vpp_if = vppif.VppApiInterface(LOG_NAME)

        self._shg = new
        for vxlan_idx in self._vpp_vxlans.values():
            self._add_ifid_to_bd( vxlan_idx, vpp_if=vpp_if, add_shg=True)

    @property
    def mcast_if(self):
        return self._mcast_if

    @mcast_if.setter
    def mcast_if(self, new):
        if_index = None
        self.logger.debug("Set MCAST IF for {} to {}".format(self._name, new))
	    
        # if new and not isinstance(new, str):
        #     raise VxLanIfErr('Bad param type expect str or None', -1)

        vpp_if = vppif.VppApiInterface(LOG_NAME)

        self._delete_vxlan_mcast_if(vpp_if=vpp_if)
        if new != None:
            if_index = self.find_mcast_if_vpp(hostifname=new, vpp_if=vpp_if)
            if if_index == None:
                raise VxLanIfErr('Cannot find interface in VPP', -1)

        self.logger.debug("Set MCAST IF for {} to {} - {}".format(self._name, new, if_index))
        self._mcast_if = if_index

    @property
    def mcast_ip(self):
        return self._mcast_ip
    
    @mcast_ip.setter
    def mcast_ip(self, new):
        self.logger.debug("Set MCAST ADDR for {} to {}".format(self._name, new))
        if new and not isinstance(new, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            raise VxLanIfErr('Bad param type expect ipaddress.IPv4Address, ipaddress.IPv6Address or None', -1)
        if new and not ( str(new).startswith('239') or str(new).startswith('ff')):
            raise VxLanIfErr('Bad IP address expect multicast in range 239.0.0.0/8 for IPv4 and FF00::/8 for IPv6', -1)

        vpp_if = vppif.VppApiInterface(LOG_NAME)
        self._delete_vxlan_mcast_if(vpp_if=vpp_if)
        self._mcast_ip = new
        mcast_idx = self._create_vxlan_mcast_if(vpp_if=vpp_if)
        if mcast_idx and mcast_idx != 0xffffffff:
            self._add_ifid_to_bd(mcast_idx, vpp_if=vpp_if, add_shg=True)
            self._vpp_vxlans[self._mcast_ip] = mcast_idx

    def down_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifdown", "--force", self.__get_sys_ifname()])

    def up_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifup", self.__get_sys_ifname()])

    def find_mcast_if_vpp(self, hostifname, vpp_if=None):
        ifname = None
        idx = None
        bd = None
        if vpp_if == None:
            vpp_if = vppif.VppApiInterface(LOG_NAME)
        
        if hostifname.startswith('lo'):
            loop_id = int(hostifname.replace('lo','')) + self.LOOP_OFFSET
            ifname = 'loop' + str(loop_id)
            idx = vpp_if.get_ifidx_by_name(ifname)
            if ifname == None or idx == None:
                return None
            return idx
        else:
            (ifname, idx) = vpp_if.search_vppif_by_hostifname(hostifname)
            if ifname == None or idx == None:
                return None

            bd =  vpp_if.get_interface_bd(idx)
            if bd != None and bd['bvi_sw_if_index']:
                return bd['bvi_sw_if_index']
        return None        
