#!/usr/bin/env python3

import json
import logging
import logging.handlers
import ipaddress
import subprocess
from systemd.journal import JournalHandler
import vppapi_if.vppapi_if as vppif

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64

import edgility_db.edgility_db as db
from sysrepo_wg_concentrator import sysrepo_wg_concentrator as wg_api

class WireGuardConcentratorErr(Exception):
    def __init__(self, message, recover_f=lambda _: None):
        super(WireGuardConcentratorErr, self).__init__(message)

class WireGuardConcentratorClient():

    DEFAULT_KEEPALIVE_VALUE = 0
    NO_ERR_MSG = ''
    DNS_CONF = '/etc/dnsmasq.d/vpn-dns.conf'

    STATIC_POOL_NAME = 'static_clients_'
    DYNAMIC_POOL_NAME = 'dynamic_clients_'
    DYNAMIC_POOL_COLUMNS = [
            "public_key",
            "hostname",
            "peer_endpoint",
            "tunnel_ip",
            "allowed_ips",
            "password"
        ]
    STATIC_POOL_COLUMNS = [
            "hostname",
            "tunnel_ip"
    ]
    TIMESTAMP_POOL_NAME = 'timestamp'


    def __init__(self, hostname, password, public_key, conc_port, endpoint):
        self._hostname = hostname
        self._password = password
        self._public_key = public_key
        self._port = conc_port
        self._endpoint = endpoint
        self._client_allowed_ips = None
        self._client_tunnel_ip = None
        self._client_tunnel_ip_str_from_db = None

        self._client_if_index = None
        self._client_peer_idx = None

        self._conc_ifname = None
        self._if_index = None
        self._conc_tunnel_ip = None
        self._conc_allowed_ips = None
        self._net_info = None

        self._db = db.EdgilityDatabase()

        self._last_peer_added = None
        self._last_route_added = None

        self._register_in_db = True

        # Client reponse data
        self._client_data = {
            "tunnel-ip": '',
            "concentrator-public-key": '',
            "allowed-ips": '',
            "concentrator-tunnel-ip": '',
            "domain-name": ''
        }

        self._init_logger()

        # Real work happens here
        self._add_client()

    @property
    def conf_ifname(self):
        return self._conc_ifname

    def db_table_name(self, db_pool, intf_name):
            return db_pool + intf_name

    @conf_ifname.setter
    def conf_ifname(self, value):
        self._conc_ifname = value

        self._dynamic_table_name = self.db_table_name(self.DYNAMIC_POOL_NAME, self._conc_ifname)
        self._static_table_name = self.db_table_name(self.STATIC_POOL_NAME, self._conc_ifname)

    @classmethod
    def WireGuardConcentratorClientStartup(cls, wg_if_idx, conc_ifname, conc_port, ipv4_prefix):
        cself = cls.__new__(cls)
        super(WireGuardConcentratorClient, cself).__init__() 

        cself._if_index = wg_if_idx
        cself._port = conc_port
        cself.conf_ifname = conc_ifname
        cself._conc_tunnel_ip = ipv4_prefix

        cself._db = db.EdgilityDatabase()

        cself._init_logger()

        try:
            cself._load_clients_on_startup(cself._conc_tunnel_ip)
        except WireGuardConcentratorErr as ex:
            # TODO: recover here
            cself._logger.error(f'Printing exception {ex}')

    # delete single client from VPP and database
    @classmethod
    def WireGuardConcentratorDelClient(cls, int_name, client_choice, client_param):
        cself = cls.__new__(cls)
        super(WireGuardConcentratorClient, cself).__init__()

        cself._db = db.EdgilityDatabase()
        cself._init_logger()

        try:
            if client_choice == "public_key":
                client_ip = cself.get_tunnel_ip_from_peer_in_vpp_by_key(client_param)
                cself._db.delete_value_from_table(cself.db_table_name(cself.DYNAMIC_POOL_NAME, int_name), 'public_key', client_param)
                if client_ip:
                    cself._del_client_from_vpp(client_ip, cself._client_if_index, cself._client_peer_idx)
            elif client_choice == "hostname":
                pub_key = cself._db.fetch_from_table_by_param(cself.db_table_name(cself.DYNAMIC_POOL_NAME, int_name), 'public_key', 'hostname', client_param)
                cself._db.delete_value_from_table(cself.db_table_name(cself.DYNAMIC_POOL_NAME, int_name), 'public_key', client_param)
                client_ip = cself.get_tunnel_ip_from_peer_in_vpp_by_key(pub_key)
                if client_ip:
                    cself._del_client_from_vpp(client_ip, cself._client_if_index, cself._client_peer_idx)
        except WireGuardConcentratorErr as ex:
            cself._logger.error(f'Client - {client_param} does not exist!')

    # delete routes from clients on enable false or interface delete
    @classmethod
    def WireGuardConcentratorClientDelRoutes(cls, int_name, wgif_idx):
        cself = cls.__new__(cls)
        super(WireGuardConcentratorClient, cself).__init__()

        cself._db = db.EdgilityDatabase()
        cself._init_logger()

        try:
            cself.del_client_routes(int_name, wgif_idx)
        except WireGuardConcentratorErr as ex:
            cself._logger.debug(f'Printing exception {ex}')

    @classmethod   
    def WireGuardConcentratorClientDropTables(cls, int_name):
        cself = cls.__new__(cls)
        super(WireGuardConcentratorClient, cself).__init__()

        cself._db = db.EdgilityDatabase()
        cself._init_logger()

        cself.drop_client_tables(int_name)

    def del_client_routes(self, int_name, wgif_idx):
        dynamic_table_name = self.DYNAMIC_POOL_NAME + int_name
        clients = self._db.fetch_all_rows_from_table(dynamic_table_name)
        for client in clients:
            self._client_tunnel_ip = self.extract_from_db_tuple_resp(client, 'tunnel_ip')
            client_tunnel_ip_str = self.convert_ip_int_to_string(self._client_tunnel_ip)

            self._conc_allowed_ips = [f"{client_tunnel_ip_str}/32"]
            self._logger.debug(f"del_client_routes route - {f'{client_tunnel_ip_str}/32'}")
            self.add_del_peer_address_vpp_static_route(delete=True, wgif_idx=wgif_idx)

    def drop_client_tables(self, int_name):
        dynamic_table_name = self.DYNAMIC_POOL_NAME + int_name
        static_table_name = self.STATIC_POOL_NAME + int_name

        self._db.drop_table(dynamic_table_name)
        self._db.drop_table(static_table_name)
        self._db.execute_and_close(f"DELETE FROM {self.TIMESTAMP_POOL_NAME};")

        self._logger.debug(f"Clients pools dropped!")
                
    def __str__(self):
        return json.dumps(self._client_data)

    def get_vpp_client(self):
        return vppif.VppApiInterface('wireguard-concentrator')

    def _add_client(self): 
        try:
            self._fetch_wireguard_concentrator_instance()
            self._verify_client_credentials()
            self._decrypt_client_public_key()
            self._fetch_client_data()
            self._add_client_in_vpp()
            if self._register_in_db:
                self._register_client_in_database()

        except WireGuardConcentratorErr as ex:
            # TODO: recover here
            self._logger.error(f'Printing exception {ex}')

    
    def _verify_client_credentials(self):
        sr_wgc = wg_api.SysrepoWireGuardConcentrator(app_name="wireguard concentrator")
        conc_password = sr_wgc.get_wireguard_concentrator_param(self._conc_ifname, 'concentrator-password')
        if not self._password == conc_password:
            raise WireGuardConcentratorErr(f"Error: Incorrect password!")

    def _fetch_wireguard_concentrator_instance(self):
        self.conf_ifname = self._get_concentrator_ifname()
        self._get_wg_interface_index()
        self._fetch_concentrator_tunnel_ip()

    def _get_concentrator_ifname(self):
        sr_wgc = wg_api.SysrepoWireGuardConcentrator(app_name="wireguard concentrator")
        interfaces = sr_wgc.get_wireguard_concentrator_interfaces()['wg-concentrator']
        for interface in interfaces:
            if interface.get('concentrator-port', None) == self._port:
                return interface.get('name', None)
        raise WireGuardConcentratorErr(f"Error: Interfaces {interfaces}")
    
    def _fetch_concentrator_tunnel_ip(self):
        vpp = self.get_vpp_client()
        self._conc_tunnel_ip = vpp.get_interface_address_by_ifidx(self._if_index)
        if not self._conc_tunnel_ip:
            raise WireGuardConcentratorErr(f"Error: No ip on interface with name {self._conc_ifname}")

    def _get_wg_interface_index(self):
        vpp = self.get_vpp_client()
        interface = vpp.wg_get_interface_by_local_port(self._port)
        
        self._if_index = interface['sw_if_index']
        del(vpp)
        if not self._if_index:
            raise WireGuardConcentratorErr(f"Error: Not existing wireguard inteface with local_port {self._port}")
    
    def _load_clients_on_startup(self, conc_tunnel_ip):
        clients = self._db.fetch_all_rows_from_table(self._dynamic_table_name)
        for client in clients:
            self._public_key = self.extract_from_db_tuple_resp(client, 'public_key').strip()
            self._endpoint = self.extract_from_db_tuple_resp(client, 'peer_endpoint')
            self._client_tunnel_ip = self.extract_from_db_tuple_resp(client, 'tunnel_ip')
            
            client_tunnel_ip_str = self.convert_ip_int_to_string(self._client_tunnel_ip)

            self._conc_allowed_ips = [f"{client_tunnel_ip_str}/32"]

            self._net_info = ipaddress.IPv4Network(f"{conc_tunnel_ip}", strict=False)

            # add iptables rule on startup
            self._del_ip_tables_mariadb_rule()
            self._add_ip_tables_rule_for_mariadb_tunnel_network(self._net_info.compressed)
            self._add_ip_tables_mariadb_rule()

            try:
                self._add_client_in_vpp()   
            except Exception:
                pass # Ignore failed clients for now 

    def _fetch_client_data(self):
        if self._fetch_client_data_on_existing_client():
            self._update_hostname_in_database() # This handles cases where client changed his hostname
            self._update_peer_endpoint_in_database() # This handles cases where client's local endpoint changed dinamically

            #on client tunnel-ip change, del old peer and set new peer with new tunnel-ip

            client_old_ip = self.get_tunnel_ip_from_peer_in_vpp_by_key(self._public_key)
            client_old_ip_str = str(client_old_ip).split('/')
            client_old_ip_int = int(ipaddress.IPv4Address(client_old_ip_str[0]))

            try:
                client_new_ip_int = self._db.fetch_from_table_by_param(self._static_table_name, 'tunnel_ip', 'hostname', self._hostname)
            except Exception as e:
                self._logger.error(f"No static entry for this client.")

            if client_old_ip_int != client_new_ip_int and client_new_ip_int != None:
                self._del_client_from_vpp(client_old_ip, self._client_if_index, self._client_peer_idx)
                self._client_tunnel_ip = client_new_ip_int
                try:
                    self._db.update_column_value_on_matching_key(self._dynamic_table_name, 'tunnel_ip', self._client_tunnel_ip, 'public_key', self._public_key)
                except Exception as e:
                    self._logger.error(f"Client not in dynamic pool.")
                self._register_in_db = False
            else:
                raise WireGuardConcentratorErr(f"Client data fetched from existing entry")
        
        self._client_tunnel_ip = self._get_tunnel_ip_from_static_pool() or self._get_tunnel_ip_from_dynamic_pool()
        if not self._client_tunnel_ip:
            raise WireGuardConcentratorErr(f"Error: Could not allocate tunnel_ip for client")

        conc_public_key = self._encrypt_public_key(self._get_concentrator_public_key())

        if not conc_public_key:
            raise WireGuardConcentratorErr(f"Error: Could not get concentrator public key")

        client_tunnel_ip_str = self.convert_ip_int_to_string(self._client_tunnel_ip)
        self._conc_allowed_ips = [f"{client_tunnel_ip_str}/32"]

        self._del_ip_tables_mariadb_rule()
        self._add_ip_tables_rule_for_mariadb_tunnel_network(self._net_info.compressed)
        self._add_ip_tables_mariadb_rule()

        self._logger.debug(f"Client tunnel ip: {client_tunnel_ip_str}")

        self._client_data['tunnel-ip'] = f"{client_tunnel_ip_str}/32"
        self._client_data['concentrator-public-key'] = conc_public_key.decode("utf-8")
        self._client_data['allowed-ips'] = self._get_client_allowed_ips()[0]
        self._client_data['concentrator-tunnel-ip'] = str(self._conc_tunnel_ip).split('/')[0]
        self._client_data['domain-name'] = self._get_domain_name()

    def _fetch_client_data_on_existing_client(self):
        data = self._db.fetch_row_by_param(self._dynamic_table_name, 'public_key', self._public_key)
        self._logger.debug(f'Existing client with data {data}')
        if data:
            client_tunnel_ip_int = self.extract_from_db_tuple_resp(data, "tunnel_ip")
            self._client_tunnel_ip_str_from_db = self.convert_ip_int_to_string(client_tunnel_ip_int)

            self._client_data['tunnel-ip'] = f'{self._client_tunnel_ip_str_from_db}/32'
            self._client_data['concentrator-public-key'] = self._encrypt_public_key(self._get_concentrator_public_key()).decode("utf-8")
            self._client_data['allowed-ips'] = self.extract_from_db_tuple_resp(data, 'allowed_ips')
            self._client_data['concentrator-tunnel-ip'] = str(self._conc_tunnel_ip).split('/')[0]
            self._client_data['domain-name'] = self._get_domain_name()

            return True
        return False

    def _get_tunnel_ip_from_static_pool(self):
        self._client_allowed_ips = self._get_client_allowed_ips()
        return self._db.fetch_from_table_by_param(self._static_table_name, 'tunnel_ip', 'hostname', self._hostname)
        
    def _get_tunnel_ip_from_dynamic_pool(self):
        tunnel_ips = self._db.fetch_column_entries(self._dynamic_table_name, 'tunnel_ip') 
        self._logger.debug(f'Returning tunnel ip: {tunnel_ips}')
        return self._get_first_available_tunnel_ip(tunnel_ips)

    def _register_client_in_database(self):
        data_values = (str(self._public_key), str(self._hostname), str(self._endpoint), int(self._client_tunnel_ip), str(self._client_allowed_ips[0]), str(self._password))
        self._db.insert_values_in_table(self._dynamic_table_name, self.DYNAMIC_POOL_COLUMNS, data_values)

    def _update_hostname_in_database(self):
        self._logger.debug(f"{self._dynamic_table_name}, 'hostname', f'{self._hostname}', 'public_key', {self._public_key}")
        self._db.update_column_value_on_matching_key(self._dynamic_table_name, 'hostname', f"'{self._hostname}'", 'public_key', self._public_key)

    def _update_peer_endpoint_in_database(self):
        self._logger.debug(f"{self._dynamic_table_name}, 'peer_endpoint', f'{self._endpoint}', 'public_key', {self._public_key}")
        self._db.update_column_value_on_matching_key(self._dynamic_table_name, 'peer_endpoint', f"'{self._endpoint}'", 'public_key', self._public_key)

    def _get_first_available_tunnel_ip(self, ips):
        # Segment the length and check each one (for big ranges)
        self._net_info = ipaddress.IPv4Network(f"{self._conc_tunnel_ip}", strict=False)
        net_addr = int(self._net_info.network_address)
        available_hosts = self._net_info.num_addresses
        broadcast_addr = self._net_info.broadcast_address
        self._client_allowed_ips = self._get_client_allowed_ips()
        self._logger.debug(f"Allowed ips: {self._client_allowed_ips}")

        if not ips:
            ips = []

        #convert tuples to list
        ips = [ip[0] for ip in ips]
        
        ips.append(int(self._conc_tunnel_ip))
        
        ips.sort()

        last_ip_idx = -1
        last_ip = ips[last_ip_idx]

        if available_hosts <= len(ips):
            return None

        while True:
            if last_ip - net_addr > len(ips[:ips.index(last_ip)]) + 1:
                if last_ip_idx == -len(ips):
                    return net_addr + 1
                last_ip_idx -= 1
                last_ip = ips[last_ip_idx]
                continue
            
            if last_ip + 1 == broadcast_addr:
                return None

            return last_ip + 1

    def _add_client_in_vpp(self):
        vpp = self.get_vpp_client()
        client_tunnel_ip_str = str(ipaddress.ip_address(int(self._client_tunnel_ip)))
        ret, peer_idx = vpp.wg_add_peer(wgif=self._if_index, endpoint=self._endpoint, port=self._port, keepalive=self.DEFAULT_KEEPALIVE_VALUE,
                            allowed_ips=[f"{client_tunnel_ip_str}/32"], public_key_base64=self._public_key, base64=True)
        del(vpp)
        
        if ret == -116:
            self._client_data['error'] = 'Current client is already connected to concentrator!'
            raise WireGuardConcentratorErr("Peer already exists!")
        elif ret:
            raise WireGuardConcentratorErr(f"Error executing wg_add_peer: {ret}")
        
        self._last_peer_added = peer_idx

        ret = self.add_del_peer_address_vpp_static_route(delete=False)
        if ret:
            raise WireGuardConcentratorErr(f"Error executing add_del_peer_address_vpp_static_route: {ret}")
        
    def _del_client_from_vpp(self, client_tunnel_ip, client_if_index, client_peer_idx):
        vpp = self.get_vpp_client()
        self._logger.debug('_del_client_from_vpp')

        ip_netmask = client_tunnel_ip.network.prefixlen
        ret = vpp.del_ip_route('{}/{}' .format(client_tunnel_ip.ip, ip_netmask), next_hop_sw_if_index=client_if_index)
        self._logger.debug(f"del_ip_route {ret}")
        if ret:
            self._logger.error('Del ip route fail route {}/{} via {} ret={}'.format(client_tunnel_ip.ip, ip_netmask, client_if_index ,ret))

        ret = vpp.wg_del_peer(peer_idx=client_peer_idx)
        self._logger.debug(f"wg_del_peer {ret}")
        del(vpp)
        if ret:
            self._logger.error("Error deleting peer {}" .format(self._public_key))


    def add_del_peer_address_vpp_static_route(self, delete=False, wgif_idx=None):
        netmask = 32
        
        for a_ip in self._conc_allowed_ips:
            ip_addr = ipaddress.IPv4Interface(str(a_ip))
            ip_netmask = ip_addr.network.prefixlen
            self._logger.error("WireGuardVPNInterface vpp_static_route ip_addr - {} ip_netmask {} - netmask {} - delete {}" .format(ip_addr, ip_netmask, netmask, delete))
            
            if netmask == ip_netmask:            
                vpp_if = self.get_vpp_client()
                if delete == True:
                    if wgif_idx:
                        self._if_index = wgif_idx
                    ret = vpp_if.del_ip_route('{}/{}' .format(ip_addr.ip, ip_netmask), next_hop_sw_if_index=self._if_index)
                    if ret:
                        self._logger.error('Del ip route fail route {}/{} via {} ret={}'.format(ip_addr.ip, ip_netmask, self._if_index ,ret))
                    self._logger.error("WireGuardVPNInterface vpp_static_route static route for - {}/{} REMOVED" .format(ip_addr.ip, ip_netmask))
                else:
                    ret = vpp_if.add_ip_route('{}/{}' .format(ip_addr.ip, ip_netmask), next_hop_sw_if_index=self._if_index, check=False)
                    if ret:
                        self._logger.error('Add ip route fail route {}/{} via {} ret={}'.format(ip_addr.ip, ip_netmask, self._if_index, ret))
                    
                    self._last_route_added = f'{ip_addr.ip}/{ip_netmask}'

                    self._logger.error("WireGuardVPNInterface vpp_static_route static route for - {}/{} ADDED" .format(ip_addr.ip, ip_netmask))
                del(vpp_if)

                return ret

    def delete_single_vpp_route(self, route):
        vpp_if = self.get_vpp_client()
        ret = vpp_if.del_ip_route(route, next_hop_sw_if_index=self._if_index)
        if ret:
            self.logger.error(f'Error delete_single_vpp_route: {ret}')
        del(vpp_if)

    def _generate_key(self, password):
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=b'1V0eG071nP!cH',
                iterations=100000,
                backend=default_backend()
            )
            key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
            return key
    
    def _decrypt_client_public_key(self):
        cipher = Fernet(self._generate_key(self._password))
        decrypted_public_key = cipher.decrypt(self._public_key.encode()).decode()
        self._public_key = decrypted_public_key.strip()
        if not self._public_key:
            raise WireGuardConcentratorErr(f"Error testing {decrypted_public_key}")

    def _encrypt_public_key(self, public_key):
        cipher = Fernet(self._generate_key(self._password))
        encrypted_public_key = cipher.encrypt(public_key.encode())
        return encrypted_public_key

    def _get_concentrator_public_key(self):
        with open('/etc/wg/public.key', 'r') as f:
                public_key = f.read()

        return public_key

    def extract_from_db_tuple_resp(self, db_resp, value):
        index = self.DYNAMIC_POOL_COLUMNS.index(value)
        return db_resp[index]
    
    def convert_ip_int_to_string(self, ip_int):
        return str(ipaddress.ip_address(int(ip_int)))
    
    def _get_client_allowed_ips(self):
        self._net_info = ipaddress.IPv4Network(f"{self._conc_tunnel_ip}", strict=False)
        return [f'{self._net_info.network_address}/{self._net_info.prefixlen}']
    
    def _init_logger(self, log_level=logging.DEBUG):
        self._logger = logging.getLogger(name=self.__class__.__name__)
        self._logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self._logger.addHandler(syslog_hnd)
        
    def _get_domain_name(self):
        with open(self.DNS_CONF, 'r') as file:
            for line in file:
                if line.startswith("domain="):
                    return line.split('=')[1].strip()

    def _add_ip_tables_rule_for_mariadb_tunnel_network(self, network):
        try:
            self._logger.error(f"iptables rule added for {str(network)}")
            subprocess.check_output(
                ['/sbin/iptables', '-A', 'INPUT', '-p', 'tcp', '-s', str(network), '--dport', '3306', '-m', 'conntrack', '--ctstate',
                 'NEW,ESTABLISHED', '-j', 'ACCEPT'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise WireGuardConcentratorErr(f"Failed to add iptables rule to allow access to database for clients! Exception: {e}")

    def _add_ip_tables_mariadb_rule(self):
        try:
            subprocess.check_output(
                ['/sbin/iptables', '-A', 'INPUT', '-p', 'tcp', '-m', 'tcp', '--dport', '3306', '-j', 'DROP'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self._logger.error(f"Failed to add iptables rule! Exception: {e}")

    def _del_ip_tables_mariadb_rule(self):
        try:
            subprocess.check_output(
                ['/sbin/iptables', '-D', 'INPUT', '-p', 'tcp', '-m', 'tcp', '--dport', '3306', '-j', 'DROP'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self._logger.error(f"Failed to del iptables rule! Exception: {e}")

    def get_tunnel_ip_from_peer_in_vpp_by_key(self, wg_key):
        vpp = self.get_vpp_client()
        peer = vpp.wg_get_peer_by_public_key(wg_key)
        ip_addr = None

        if peer:
            allowed_ips = peer['allowed_ips']
            self._client_if_index = peer['wgif']
            self._client_peer_idx = peer['peer_idx']
            for a_ip in allowed_ips:
                ip_addr = ipaddress.IPv4Interface(str(a_ip))
                break

        return ip_addr

