#!/usr/bin/env python3
import os
import logging
import logging.handlers
import vppapi_if.vppapi_if as vppif
import ipaddress
import debinterface
import tempfile
import time
import random
import sys
import fileinput
from  pyroute2 import IPRoute
import subprocess
from sysrepo_interfaces import configObjNet
import json
import requests
import edgility_db.edgility_db as db
import ucpeif.wg_concentratorif as wgc
from tveapi.common import utils
import psutil
import socket
from ipaddress import IPv4Network, ip_address
import netifaces

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64

class WireGuardIfErr(Exception):
    def __init__(self, message, code):
        super(WireGuardIfErr, self).__init__(message)
        self.code = code

class WireGuardPeer(object):
    def __init__(self, wg_name, r_endpoint, logger=None, log_level=logging.DEBUG, r_port=None,
                            allowed_ips=None, persistent_keepalive=0, ipv6=False, conc_pub_key=None, is_concentrator=False):

        self.LOG_NAME = 'wg-cl-cls'
        self.logger = None
        self._peer_idx = None 
        self._r_public = conc_pub_key 
        self._allowed_ips = []
        self._allowed_ips_srting = None
        self._is_concentrator = is_concentrator

        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level ) 

        self._wg_name = wg_name
        self._r_endpoint = r_endpoint
        self._r_port = r_port
        self._persistent_keepalive = persistent_keepalive

        r_endpoint = ipaddress.ip_address(self._r_endpoint)
        if isinstance(r_endpoint, ipaddress.IPv6Address):
            self._peer_is_ipv6 = True
        else:
            self._peer_is_ipv6 = False

        self._get_peer_public_key()
        self._add_allowed_ips(allowed_ips)
            
        self.logger.error('WireGuardPeer adding WireGuard config for peer to \'{}\' - remote endpoint {} port {} allowed_ips {} keepalive {}'.format(self._wg_name, self._r_endpoint, self._r_port, self._allowed_ips, self._persistent_keepalive))

    def __del__(self):
        self.logger.error("WireGuardPeer {} - {}  __del__".format(self._wg_name, self._r_endpoint))
        if os.path.exists('/etc/wg/{}/private.key'.format(str(self._r_endpoint))):
            os.remove('/etc/wg/{}/private.key'.format(str(self._r_endpoint)))
        if os.path.exists('/etc/wg/{}/public.key'.format(str(self._r_endpoint))):
            os.remove('/etc/wg/{}/public.key'.format(str(self._r_endpoint)))
        if os.path.exists('/etc/wg/{}'.format(str(self._r_endpoint))):
            os.rmdir('/etc/wg/{}'.format(str(self._r_endpoint)))
        # call vpp to remove wg tunnel

    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=self.LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)

    @property
    def wg_name(self):
        return self._wg_name

    @property
    def r_endpoint(self):
        return self._r_endpoint

    @property
    def r_port(self):
        return self._r_port

    @r_port.setter
    def r_port(self, new):
        if new and not isinstance(new, int):
            raise WireGuardIfErr('Bad peer port param type', -1)
        
        # del
        self._r_port = new
        # create
        self.logger.error("WireGuardPeer r_port {} ".format(self._r_port))

    @property
    def r_public(self):
        return self._r_public

    @r_public.setter
    def r_public(self, new):
        if new and not isinstance(new, str):
            raise WireGuardIfErr('Bad peer public key param type', -1)
        
        # del
        self._r_public = new
        # create
        self.logger.error("WireGuardPeer r_public {} ".format(self._r_public))

    @property
    def allowed_ips(self):
        return self._allowed_ips

    @allowed_ips.setter
    def allowed_ips(self, new):
        if new and not isinstance(new, str):
            raise WireGuardIfErr('Bad allowed ips param type', -1)
        
        if self._allowed_ips_srting == new:
            return
        
        # del
        self._add_allowed_ips(new)
        self.logger.error("WireGuardPeer allowed_ips {} ".format(self._allowed_ips))

    @property
    def persistent_keepalive(self):
        return self._persistent_keepalive

    @persistent_keepalive.setter
    def persistent_keepalive(self, new):
        if new and not isinstance(new, int):
            raise WireGuardIfErr('Bad persistent_keepalive param type', -1)
        
        # del
        self._persistent_keepalive = new
        # create
        self.logger.error("WireGuardPeer persistent_keepalive {} ".format(self._persistent_keepalive))

    def _get_peer_public_key(self):
        if self._r_public:
            return
        
        filename = '/etc/wg/{}/public.key' .format(str(self._r_endpoint))
        if os.path.exists(filename):
            self.logger.error("WireGuardPeer file {} ".format(filename))
            with open(filename, 'r') as f:
                self._r_public = f.read()
        else:
            # send request to get the key from the peer
            self._r_public = self.get_r_endpoint_public_key()
            if not self._r_public:
                return
            if self._r_public == 'Unknown':
                self._r_public = None
                return
            if not os.path.exists('/etc/wg/{}' .format(str(self._r_endpoint))):
                os.mkdir('/etc/wg/{}' .format(str(self._r_endpoint)))
            with open(filename, 'w+') as public_file:
                public_file.write(str(self._r_public))
        self.logger.error("WireGuardPeer self._r_public {} ".format(self._r_public))

    def _add_allowed_ips(self, allowed_ips):
        if allowed_ips == None:
            self._allowed_ips = []
            self._allowed_ips_srting = None
            return

        ips_list = allowed_ips.split(',')
        self._allowed_ips = []
        self.logger.error("WireGuardPeer list {} ".format(ips_list))
        for i in ips_list:
            self.logger.error("WireGuardPeer list {} ".format(i))
            if i in self._allowed_ips:
                continue
            else:
                self._allowed_ips.append(i)

        self._allowed_ips_srting = allowed_ips
    
    @staticmethod
    def get_auth_token_tveapi_r_endpoint(is_ipv6, endpoint, logger):
        cfg_obj = configObjNet.ConfigObjNet("/root/local-openrc.env")
        # cfg_obj = configObjNet.ConfigObjNet("/root/admin-openrc.env")
        if cfg_obj and cfg_obj['OS_USERNAME'] and cfg_obj['OS_USERNAME'] != "":
            user = cfg_obj['OS_USERNAME']
        else:
            user = "admin"
        if cfg_obj and cfg_obj['OS_PASSWORD'] and cfg_obj['OS_PASSWORD'] != "":
            password = cfg_obj['OS_PASSWORD']
        else:
            password = "admin"

        if is_ipv6:
            url = "https://[%s]:7374/v3/auth/tokens" % endpoint
        else:
            url = "https://%s:7374/v3/auth/tokens" % endpoint
        headers = {'Content-type': 'application/json', }
        data = '{"user":"%s","password":"%s"}' % (user, password)
        
        for attemt in range(2):
            try:
                response = requests.get(url, headers=headers, data=data, verify=False, auth=(user, password),timeout=10)
                return response.text
            except requests.ConnectionError:
                logger.error("WireGuardPeer Connection to \"%s\" FAILED!" % url)
                time.sleep(4)
            except requests.Timeout:
                logger.error("WireGuardPeer Connection to \"%s\" TIMED OUT!" % url)
                time.sleep(4)
            except:
                logger.error("WireGuardPeer Failed to get authentication token!")
                time.sleep(4)
        else:
            logger.error("WireGuardPeer Failed to get authentication token 2 attempts!")
        return None

    def get_r_endpoint_public_key(self):
        if self._r_public:
            return self._r_public

        token = None
        token = self.get_auth_token_tveapi_r_endpoint(self._peer_is_ipv6, self._r_endpoint, self.logger)
        if token is None:
            return None

        self.logger.error('going to send request for peer - {} - public key'.format(self._r_endpoint))
        if self._peer_is_ipv6:
            url = "https://[%s]:7373/v1/wireguard/public-key" % self._r_endpoint
        else:
            url = "https://%s:7373/v1/wireguard/public-key" % self._r_endpoint
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }
        # params = (('public-key', public_key),)
        try:
            response = requests.get(url, headers=headers, verify=False, timeout=10)
            self.logger.error("WireGuardPeer get_r_endpoint_public_key response: %s" % response.text)            
            return response.json().get('public-key')
        except:
            self.logger.error("WireGuardPeer Error occurred while getting peer - {} - public key." .format(self._r_endpoint))
        return None

class WireGuardVPNInterface(object):
    WG_OFFSET=900
    CFG_FILE_PATH='/etc/network/interfaces.d/ifcfg-{}'
    INTERFACES_FILE='/etc/network/interfaces'
    DNS_CONF_FILE='/etc/dnsmasq.d/vpn-dns.conf'
    DNSMASQ_HOSTS_DIR = '/config/dnsmasq/hostsdir'

    DYNAMIC_POOL_PROPERTIES = (
            "public_key VARCHAR(255) UNIQUE, "
            "hostname VARCHAR(255), "
            "peer_endpoint VARCHAR(255), "
            "tunnel_ip INTEGER, "
            "allowed_ips VARCHAR(255), "
            "password VARCHAR(255)"
        )
    STATIC_POOL_PROPERTIES = (
            "hostname VARCHAR(255) UNIQUE, "
            "tunnel_ip VARCHAR(255)"
        )

    TIMESTAMP_POOL_NAME = 'timestamp'
    TIMESTAMP_POOL_PROPERTIES = (
            "public_key VARCHAR(255) UNIQUE, "
            "serial_number VARCHAR(255), "
            "updated_date TIMESTAMP"
    )

    def __init__(self, name, logger=None, log_level=logging.DEBUG, enable=False, l_port=None, l_endpoint=None, l_endpoint_v6=None,
                            l_mtu=1420, ip4_addr=None, ip6_addr=None, ipv6=False,
                            conc_endpoint=None, conc_password=None, conc_port=None, is_concentrator=False,
                            d_lookup=False, d_name='edgilityconcentrator.com', a_list={}, wg_count=0):

        self.LOG_NAME = 'wg-cl-cls'
        self.logger = None
        self._wgif_idx = None
        self._peers = []
        self._enable = False
        self._l_public = None
        self._l_private = None

        if logger:
            self.logger = logger
        else:
            self.__init_logger( log_level ) 

        self._name = name
        self._l_port = l_port
        self._l_endpoint = l_endpoint
        self._l_endpoint_v6 = l_endpoint_v6
        self._mtu = l_mtu
        self._enable = enable
        
        self._conc_endpoint = self.__get_ip_from_hostname(conc_endpoint)
        self._conc_password = conc_password
        self._conc_port = conc_port
        self._conc_pub_key = None
        self._d_lookup = d_lookup
        self._d_name = d_name
        self._a_list = a_list

        self._tapidx = None
        self._ip4_addr = ip4_addr
        self._ip6_addr = ip6_addr
        self._ipv6 = ipv6
        self._mac = self.__generate_mac()

        self._is_concentrator = is_concentrator
        self._client_allowed_ips = None
        self._conc_tunnel_ip = None
        self._default_acl_rule_idx = False

        # If not explicitly set by the user, we configure local port and endpoint
        self.__configure_local_endpoint_and_port(wg_count)

        if self._l_port and (self._l_endpoint or self._l_endpoint_v6):
            self._generate_keys()
        self._create_wg_tapif()
        self.__update_cfg_file()

        try:
            subprocess.check_call(['ip', 'link', 'set', 'dev', self.__get_sys_ifname(), 'arp', 'off'])
        except subprocess.CalledProcessError as e:
            self.logger.error('Failed to disable ARP for WireGuard interface {}' .format(self.__get_sys_ifname()))

        self.down_sys_if()

        self.create_wg_interface()
        if self._enable:
            self.up_sys_if()
            try:
                self.set_host_mac_equal_to_tap_mac()
            except Exception as e:
                self.logger.debug(f"No tap interface: {e}")
        self.logger.error('WireGuardVPNInterface adding WireGuard config for \'{}\' - local endpoint {} port {} - enabled {}'.format(self._name, self._l_endpoint, self._l_port, self._enable))

        if not self.set_wireguard_concentrator() and not self._is_concentrator:
            return

        self.logger.debug(f"self._is_concentrator {self._is_concentrator}")

        if self._is_concentrator:
            self._edge_db = db.EdgilityDatabase()
            self._edge_db.create_table('dynamic_clients_' + self._name, self.DYNAMIC_POOL_PROPERTIES)
            self._edge_db.create_table('static_clients_' + self._name, self.STATIC_POOL_PROPERTIES)

            if not self.verify_if_table_exists_in_db(self.TIMESTAMP_POOL_NAME):
                self._edge_db.create_table(self.TIMESTAMP_POOL_NAME, self.TIMESTAMP_POOL_PROPERTIES)
                self.create_wireguard_keepalive_user()

            self._dnsmasq_add_del_listen_interface(self._name)
            #enable DNS service and disable systemd-resolv
            self._enable_dnsmasq()

            #start clean clients service
            os.system("/scripts/wg_client_cliener.py &")

    def __del__(self):
        self.logger.error("WireGuardVPNInterface {} __del__".format(self._name))

        interfaces = debinterface.Interfaces(interfaces_path=self.__get_cfg_fname())
        interfaces.downAdapter(self.__get_sys_ifname())

        os.unlink(self.__get_cfg_fname())

        # del tunnel interface
                
        self.delete_wg_interface()
        self._delete_wg_tapif()
        for peer in self._peers:
            if self._conc_endpoint:
                #stop client process here
                try:
                    pid = self.get_pid_by_command(f"python3 /scripts/wg_client_thread.py {self._name} {self._conc_tunnel_ip} {self._l_public}")
                    self.logger.error(f"PESHO2 pid {pid}")
                    os.kill(int(pid), 9)
                    self.logger.error(f"Wireguard keepalive process killed! {self._name} {self._conc_tunnel_ip} {self._l_public}")
                except Exception as e:
                    self.logger.error("Wireguard client process does not exist.")
                self.del_routes_from_concentrator(peer)
            self._peers.remove(peer)
            del peer

        if self._is_concentrator:
            try:
                self._del_ip_tables_mariadb_rule()
                if self._ip4_addr:
                    network = ipaddress.ip_network(self._ip4_addr, strict=False)

                    subprocess.check_output(
                        ['/sbin/iptables', '-D', 'INPUT', '-p', 'tcp', '-s', str(network.compressed), '--dport', '3306', '-m', 'conntrack', '--ctstate',
                            'NEW,ESTABLISHED', '-j', 'ACCEPT'], stderr=subprocess.STDOUT)
                self._add_ip_tables_mariadb_rule()
            except subprocess.CalledProcessError as e:
                self.logger.error(f"Failed to delete iptables rule that allows access to database for clients! Exception: {e}")

            try:
                pid = self.get_pid_by_command('python3 /scripts/wg_client_cliener.py')
                os.kill(int(pid), 9)
            except Exception as e:
                self.logger.error(f"There is no running cleaner process.")

            # get all ACLs with tag wg_if_name and del them from VPP
            acl_idx = self._search_for_acl_by_tag(self._name)
            for acl in acl_idx:
                self._del_acl(acl)

            self._dnsmasq_add_del_listen_interface(self._name, delete=True)
            utils.execute_shell_command(f'sudo systemctl restart dnsmasq.service')

            # drop clients from tables 
            wgc.WireGuardConcentratorClient.WireGuardConcentratorClientDropTables(self.name)

    def __init_logger(self, log_level ):
        self.logger = logging.getLogger(name=self.LOG_NAME)
        self.logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self.logger.addHandler(syslog_hnd)

    def __generate_mac(self):
        macarr =  [0xfe, random.getrandbits(8),
           random.getrandbits(8), random.getrandbits(8),
           random.getrandbits(8), random.getrandbits(8)]
               
        return ':'.join(["%02x" % x for x in macarr])
    
    def __configure_local_endpoint_and_port(self, wg_count):
        if not self._l_port:
            self._l_port = int(self._name[2:]) # removes 'wg' from name
        if not self._l_endpoint or not self._l_endpoint_v6:
            self._l_endpoint = self._get_local_endpoint()

    def get_ip_address(self, ifname, is_v6=False):
        try:
            return netifaces.ifaddresses(ifname)[10 if is_v6 else 2][0]['addr']
        except:
            return ''

    def _get_local_endpoint(self):
        iface_name = ""
        ifacev6_name = ""

        for i in range(0 , 2):
            lowest_metric = sys.maxsize
            if self._conc_endpoint:
                # TODO: ihandjiiski : use pyroute2 instead of subprocess
                ip_routes_t = subprocess.Popen(['ip', '-6' if i == 0 else '-4', 'route', 'get', self._conc_endpoint],
                                                stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).communicate()
            else:
                ip_routes_t = subprocess.Popen(['ip', '-6' if i == 0 else '-4', 'route', 'show', 'default'],
                                               stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).communicate()
            ip_routes_l = list(ip_routes_t)
            ip_routes_l = [x for x in ip_routes_l if x is not None and x != b'']
            if not ip_routes_l:
                continue
            for route in ip_routes_l:
                info = route.decode("utf-8").split()
                try:
                    metric_idx = info.index('metric') + 1
                    metric = int(info[metric_idx])
                except ValueError:
                    metric = 0
                iface_idx = info.index('dev') + 1
                if metric < lowest_metric:
                    lowest_metric = metric
                    if (i == 0): # ipv6
                        ifacev6_name = info[iface_idx]
                    else:
                        iface_name = info[iface_idx]

        # prioritize ipv4 for now and revisit for ipv6
        return self.get_ip_address(iface_name or ifacev6_name)

    @property
    def conc_endpoint(self):
        return self._conc_endpoint
    
    @conc_endpoint.setter
    def conc_endpoint(self, new):
        if new and not isinstance(new, ipaddress.IPv4Address):
            raise WireGuardIfErr('Bad Concentrator endpoint param type expect ipaddress.IPv4Address or None', -1)
        
        if self._conc_endpoint == new:
            return

        self._conc_endpoint = new
        for peer in self._peers:
            if self._conc_endpoint:
                self.del_routes_from_concentrator(peer)
            self._peers.remove(peer)
            del peer

    @property
    def conc_password(self):
        return self._conc_password
    
    @conc_password.setter
    def conc_password(self, new):
        if new and not isinstance(new, str):
            raise WireGuardIfErr('Bad local port param type', -1)
        if self._conc_password == new:
            return

        self._conc_password = new
        for peer in self._peers:
            if self._conc_endpoint:
                self.del_routes_from_concentrator(peer)
            self._peers.remove(peer)
            del peer
    
    @property
    def conc_port(self):
        return self._conc_port
    
    @conc_port.setter
    def conc_port(self, new):
        if new and not isinstance(new, int):
            raise WireGuardIfErr('Bad Concentrator port param type', -1)
        
        if self._conc_port == new:
            return

        self._conc_port = new
        for peer in self._peers:
            if self._conc_endpoint:
                self.del_routes_from_concentrator(peer)
            self._peers.remove(peer)
            del peer

    @property
    def name(self):
        return self._name

    @property
    def l_port(self):
        return self._l_port

    @l_port.setter
    def l_port(self, new):
        if new and not isinstance(new, int):
            raise WireGuardIfErr('Bad local port param type', -1)
        
        if self._l_port == new:
            return

        self.delete_wg_interface()
        self._l_port = new
        if self._l_port and (self._l_endpoint or self._l_endpoint_v6):
            self._generate_keys()
        self.create_wg_interface()

        if self._is_concentrator and self._wgif_idx:
            self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))

        self.logger.error("WireGuardVPNInterface l_port {} ".format(self._l_port))

    @property
    def l_endpoint(self):
        return self._l_endpoint

    @l_endpoint.setter
    def l_endpoint(self, new):
        if new and not isinstance(new, ipaddress.IPv4Address):
            raise WireGuardIfErr('Bad local endpoint param type expect ipaddress.IPv4Address or None', -1)
        
        if self._l_endpoint == new:
            return

        # delete old keys
        
        self.delete_wg_interface()
        self._l_endpoint = new
        if self._l_port and (self._l_endpoint or self._l_endpoint_v6):
            self._generate_keys()
        self.create_wg_interface()

        if self._is_concentrator and self._wgif_idx:
            self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))

        self.logger.error("WireGuardVPNInterface l_endpoint {} ".format(self._l_endpoint))

    @property
    def l_endpoint_v6(self):
        return self._l_endpoint_v6

    @l_endpoint_v6.setter
    def l_endpoint_v6(self, new):
        if new and not isinstance(new, ipaddress.IPv6Address):
            raise WireGuardIfErr('Bad local ipv6 endpoint param type expect ipaddress.IPv6Address or None', -1)
        
        if self._l_endpoint_v6 == new:
            return

        self.delete_wg_interface()
        self._l_endpoint_v6 = new
        if self._l_port and (self._l_endpoint or self._l_endpoint_v6):
            self._generate_keys()
        self.create_wg_interface()
        self.logger.error("WireGuardVPNInterface l_endpoint_v6 {} ".format(self._l_endpoint_v6))

    @property
    def l_public(self):
        return self._l_public

    @l_public.setter
    def l_public(self, new):
        if new and not isinstance(new, str):
            raise WireGuardIfErr('Bad local public key param type', -1)

        if self._l_public == new:
            return

        self.delete_wg_interface()
        self._l_public = new
        self.create_wg_interface()

        if self._is_concentrator and self._wgif_idx:
            self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))

        self.logger.error("WireGuardVPNInterface l_public {} ".format(self._l_public))

    @property
    def l_private(self):
        return self._l_private

    @l_private.setter
    def l_private(self, new):
        if new and not isinstance(new, str):
            raise WireGuardIfErr('Bad local private key param type', -1)

        if self._l_private == new:
            return

        self.delete_wg_interface()
        self._l_private = new
        self.create_wg_interface()

        if self._is_concentrator and self._wgif_idx:
            self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))

        self.logger.error("WireGuardVPNInterface l_private {} ".format(self._l_private))

    @property
    def ip4_addr(self):
        return self._ip4_addr

    @ip4_addr.setter
    def ip4_addr(self, new):
        if new and not isinstance(new, ipaddress.IPv4Interface):
            raise WireGuardIfErr('Bad ipv4 prefix key param type expect ipaddress.IPv4Address or None', -1)

        if self._ip4_addr == new:
            self._add_concentrator_routes()
            return

        if self._enable:
            self.down_sys_if()

        # del ip address from vpp interface
        if self._wgif_idx and self._ip4_addr:
            vpp_if = vppif.VppApiInterface(self.LOG_NAME)
            if self._tapidx:
                ret = vpp_if.drctl_set_ip(self._tapidx, self._ip4_addr.ip, delete=1 )
                self.logger.error("WireGuardVPNInterface delete address {} to wireguard interface" .format(str(self._ip4_addr.with_netmask)))
            ret = vpp_if.del_interface_address(self._wgif_idx, str(self._ip4_addr.with_netmask))
            if ret:
                self.logger.error("WireGuardVPNInterface error deleting address from wireguard interface")
            del(vpp_if)
                
        self._ip4_addr = new
        self._update_vpp_ip_address()
            
        self.__update_cfg_file()
        
        if self._a_list:
            self._process_acl_rules(self._a_list)

        if self._is_concentrator and self._wgif_idx:
            self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))

        if self._enable:
            self.up_sys_if()
            try:
                self.set_host_mac_equal_to_tap_mac()
            except Exception as e:
                self.logger.debug(f"No tap interface: {e}")
            self._add_concentrator_routes()
            
        self.logger.error("WireGuardVPNInterface ip4_addr {} ".format(self._ip4_addr))

    @property
    def ip6_addr(self):
        return self._ip6_addr

    @ip6_addr.setter
    def ip6_addr(self, new):
        if new and not isinstance(new, ipaddress.IPv6Interface):
            raise WireGuardIfErr('Bad ipv4 prefix key param type expect ipaddress.IPv6Address or None', -1)

        if self._ip6_addr == new:
            return

        if self._enable:
            self.down_sys_if()

        self._ip6_addr = new
        
        self.__update_cfg_file()

        if self._enable:
            self.up_sys_if()
        self.logger.error("WireGuardVPNInterface ip6_addr {} ".format(self._ip6_addr))

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, new):
        if new and not isinstance(new, bool):
            raise WireGuardIfErr('Bad enable param type', -1)

        if self._enable == new:
            return

        if self._enable:
            self.down_sys_if()

        self.delete_wg_interface()
        self._enable = new
        self.create_wg_interface()
        self.__update_cfg_file()

        if new:
            self.up_sys_if()
        
        if new == True:
            self.logger.error(f"set_wireguard_concentrator {new}")
            try:
                self.set_wireguard_concentrator()
            except WireGuardIfErr as e:
                self.logger.error(f"Failed to set WireGuard concentrator: {e}")
                return
            if self._is_concentrator and self._wgif_idx:
                self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))

        self.logger.error("WireGuardVPNInterface enable {} ".format(self._enable))
        
    @property
    def d_lookup(self):
        return self._d_lookup
    
    @d_lookup.setter
    def d_lookup(self, new):
        if new and not isinstance(new, bool):
            raise WireGuardIfErr('Bad domain-lookup param type', -1)
        
        self._d_lookup = new
        
        if self._enable:
            self._set_domain_lookup(new)
        
        self.logger.error("WireGuardVPNInterface domain-lookup {} ".format(new))
        
    @property
    def d_name(self):
        return self._d_name
    
    @d_name.setter
    def d_name(self, new):
        
        if new and not isinstance(new, str):
            raise WireGuardIfErr('Bad domain-name param type', -1)
        
        self._set_domain_name(new)
        self.logger.error("WireGuardVPNInterface domain-name {} ".format(self._d_name))

    @property
    def a_list(self):
        return self._a_list

    @a_list.setter
    def a_list(self, new):

        if not isinstance(new, dict):
            raise WireGuardIfErr('ACL list should be a dictionary.', -1)

        self._a_list = new

        # Check if the ACL rule contains 'delete' in any of its values and process the deletion
        if self._process_acl_deletion(new):
            return

        # If no 'delete' keyword was found, process the ACL normally
        self._process_acl_rules(new)

    @property
    def mtu(self):
        return self._mtu

    @mtu.setter
    def mtu(self, new):
        if new and not isinstance(new, int):
            raise WireGuardIfErr('Bad MTU param type', -1)
        
        if self._mtu == new:
            return

        self._mtu = new
        self._set_interface_mtu()
        self.logger.error("WireGuardVPNInterface mtu {} ".format(self._mtu)) 

    # here we configure a concentrator as peer for our clients
    def set_wireguard_concentrator(self):
        if self.if_wg_concentrator_valid_prerequisites():
            data = self._get_wireguard_data()

            if not data:
                self.logger.error('Client already connected to Wireguard Concentrator!')
                return

            try:
                self._conc_tunnel_ip = data['concentrator-tunnel-ip']
            except:
                self.logger.error('Connecting to Wireguard Concentrator failed!')
                return

            if not data['tunnel-ip']:
                self.logger.error("Failed to obtain tunnel IP from Wireguard Concentrator! Please verify concentrator parameters!", -1)
                return

            self._conc_pub_key = data['concentrator-public-key']
            self._client_allowed_ips = data['allowed-ips']
            self._decrypt_public_key()

            self.init_peer(self._conc_endpoint, self._conc_port, self._client_allowed_ips, 1, conc_pub_key=self._conc_pub_key, is_concentrator=True)

            self.ip4_addr = ipaddress.IPv4Interface(data['tunnel-ip'])
            self._set_domain_lookup(self._d_lookup)
            self._d_name = data['domain-name']
            self._set_conc_domain_locally()
            self._set_domain_name(self._d_name)

    def if_wg_concentrator_valid_prerequisites(self):
        local_endpoint = self._l_endpoint or self._l_endpoint_v6
        prerequisites = (self._enable, self._l_port, local_endpoint, self._conc_endpoint
                                               , self._conc_password, self._conc_port)

        if not [x for x in prerequisites if x is None or x == '' or x is False]:
            return True
        return False
    
    def init_peer(self, r_endpoint, r_port, allowed_ips=None, persistent_keepalive=0, conc_pub_key=None, is_concentrator=True):
        self.logger.error("WireGuardVPNInterface init_peer peer {} port {} allowed_ips {} keepalive {}".format(r_endpoint, r_port, allowed_ips, persistent_keepalive))
        peer = next(( a for a in self._peers if a._r_endpoint == r_endpoint), None)
        if peer:
            self.logger.error("WireGuardVPNInterface peer \'{}\' allready exists.".format(peer._r_endpoint))
        else:
            self.logger.error("Add WireGuard peer configuration for \'{}\' interface \'{}\'" .format(r_endpoint, self._name))
            peer = WireGuardPeer(wg_name=self._name, r_endpoint=r_endpoint, logger=self.logger, log_level=logging.DEBUG, r_port=r_port,
                            allowed_ips=allowed_ips, persistent_keepalive=persistent_keepalive, conc_pub_key=conc_pub_key, is_concentrator=is_concentrator)
            self._peers.append(peer)
            self.add_peer(peer)

    def add_peer_param(self, r_endpoint, param=None, value=None):
        self.logger.error("WireGuardVPNInterface add_peer_param peer {} param {} value {}".format(r_endpoint, param, value))
        peer = next(( a for a in self._peers if a._r_endpoint == r_endpoint), None)
        if peer:
            self.logger.error("WireGuardVPNInterface peer \'{}\' allready exists.".format(peer._r_endpoint))
        else:
            if value:
                self.logger.error("Del WireGuard peer configuration for \'{}\' interface \'{}\'" .format(r_endpoint, self._name))
                peer = WireGuardPeer(wg_name=self._name, r_endpoint=r_endpoint, logger=self.logger, log_level=logging.DEBUG, r_port=None,
                                allowed_ips=None, persistent_keepalive=None)
                self._peers.append(peer)
        
        self.delete_peer(peer)
        if param and value:
            self.logger.error("WireGuardVPNInterface peer param {}, value {}" .format(param, value))            
            peer.param = value
            if "r_port" == param:
                peer.r_port = int(value)
            elif "allowed_ips" == param:
                peer.allowed_ips = value
            elif "persistent_keepalive" == param:
                peer.persistent_keepalive = int(value)
            self.add_peer(peer)

    def del_peer_param(self, r_endpoint, param=None, value=None):
        self.logger.error("WireGuardVPNInterface del_peer_param peer {} param {} value {}".format(r_endpoint, param, value))
        peer = next(( a for a in self._peers if a._r_endpoint == r_endpoint), None)
        if not peer:
            self.logger.error("WireGuardVPNInterface peer \'{}\' does not exist.".format(peer._r_endpoint))
        else:
            self.logger.error("Del WireGuard peer configuration for \'{}\' interface \'{}\'" .format(r_endpoint, self._name))
            
            self.delete_peer(peer)
            if param:
                peer.param = value
                self.add_peer(peer)
            else:
                # peer is deleted
                self._peers.remove(peer)
                del peer

    def _set_interface_mtu(self):
        """ Need to configure MTU on both 
            - wgXX interface in VPP
            - host repersentation in user space
        """
        if not self._mtu:
            self.logger.error('MTU is not configured')
            return

        with IPRoute() as ipr:
            index = ipr.link_lookup(ifname=self.__get_sys_ifname())[0]
            ipr.link('set', index=index, mtu=self._mtu)

        if self._wgif_idx:
            vpp_if = vppif.VppApiInterface(self.LOG_NAME)
            ret = vpp_if.set_interface_mtu(self._wgif_idx, self._mtu)
            self.logger.error('set_interface_mtu ret {}'.format(ret))
            del(vpp_if)

    def _create_wg_tapif(self):
        if not self._name:
            self.logger.error('Interface name is not set!')
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        #get tap ind
        tap_id = int(self._name.replace('wg',''))
        self.logger.error('wg ID {}'.format(tap_id))

        (ret, tapidx)=vpp_if.create_tap_interface(self.__get_sys_ifname(), id=tap_id+self.WG_OFFSET, 
                                                            host_mac_addr=self._mac, 
                                                            mac_addr=self._mac,
                                                            host_mac_addr_set=True,
                                                            host_mtu_size=self._mtu)

        if 0 != ret:
            self.logger.error('create_tap_interface error - {}' .format(ret))
            return

        #Create bridge-domain
        ret = vpp_if.create_bridge_domain(tapidx)
        #Attach interface to bridge-domain
        vpp_if.add_to_bridge(tapidx,tapidx)
        # configure drctl and l2 emulation
        vpp_if.set_interface_learning(tapidx, disable=1)
        # setting l2 emulation on interface returns always "unpack_from requires a buffer of at least 18 bytes for unpacking 8 bytes at offset 10 (actual buffer size is 10)"
        try:
            ret = vpp_if.l2_emulation_enable_disable(tapidx, enable=1)
        except Exception as e:
            self.logger.debug(f"L2 emulation set on {tapidx}. Disregard exception: {e}")
        vpp_if.enable_drctl(tapidx, mac=self._mac)
        vpp_if.ifup(tapidx)

        self._tapidx = tapidx
        del(vpp_if)

    def _delete_wg_tapif(self):
        if not self._tapidx:
            self.logger.error('Interface does not exists!')
            return
        
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        vpp_if.ifdown(self._tapidx)
        vpp_if.disable_drctl(self._tapidx, self._mac)
        # deleting l2 emulation from interface returns always "unpack_from requires a buffer of at least 18 bytes for unpacking 8 bytes at offset 10 (actual buffer size is 10)"
        try:
            ret = vpp_if.l2_emulation_enable_disable(self._tapidx, enable=0)
        except Exception as e:
            self.logger.debug(f"L2 emulation deleted from {self._tapidx}. Disregard exception: {e}")
        vpp_if.delete_tap_interface(self._tapidx)

        # cleanup ARP term entries
        # vpp_if.flush_bd_ip_mac_entries(bd_id=self._tapidx)

        self._tapidx = None
        del(vpp_if)

    def __get_sys_ifname(self):
        # No 'br-' prefix
        return self._name

    def __get_cfg_fname(self):
        return self.CFG_FILE_PATH.format(self.__get_sys_ifname())

    def down_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifdown", "--force", self.__get_sys_ifname()])

    def up_sys_if(self):
        debinterface.toolutils.safe_subprocess(["/sbin/ifup", "--force", self.__get_sys_ifname()])

    def __update_cfg_file(self):
        sys_ifname=self.__get_sys_ifname()
        interfaces_path=self.__get_cfg_fname()

        if not os.path.exists(interfaces_path):
            with open(interfaces_path, 'w') as f:
                pass
            self.update_interfaces_file()

        (fh, backup) = tempfile.mkstemp(dir='/tmp')
        os.close(fh)

        interfaces = debinterface.Interfaces(interfaces_path=interfaces_path, backup_path=backup)
        interfaces.removeAdapterByName(sys_ifname)

        ipv4_set = self.__get_ipv4_adapter_cfg()
        ipv6_set = self.__get_ipv6_adapter_cfg()

        adapter = ipv4_set if ipv4_set else ipv6_set

        adapter.setAuto(self._enable)
        adapter.setHotplug(True)

        index = 0
        if ipv4_set:
            interfaces.addAdapter( ipv4_set.export(), index)
            index += 1
        
        if ipv6_set:
            interfaces.addAdapter( ipv6_set.export(), index)

        interfaces.writeInterfaces()

        os.unlink(backup)

    def __get_ipv4_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv4_set = debinterface.NetworkAdapter(sys_ifname)
        ipv4_set.setName(sys_ifname)
        ipv4_set.setAddrFam('inet')

        if self._ip4_addr:
            ipv4_set.setAddressSource('static')
            ipv4_set.setAddress(str(self._ip4_addr.ip))
            ipv4_set.setNetmask(str(self._ip4_addr.network.netmask))
        else:
            ipv4_set.setAddressSource('manual')

        return ipv4_set
        
    def __get_ipv6_adapter_cfg(self):
        sys_ifname=self.__get_sys_ifname()
        ipv6_set = debinterface.NetworkAdapter(sys_ifname)
        ipv6_set.setName(sys_ifname)
        ipv6_set.setAddrFam('inet6')

        if self._ip6_addr:
            ipv6_set.setAddressSource('static')
            ipv6_set.setAddress(str(self._ip6_addr.ip))
            ipv6_set.setNetmask(str(self._ip6_addr.network.prefixlen))
        else:
            return None

        ipv6_set.appendPreUp('echo 0 > /proc/sys/net/ipv6/conf/{}/accept_dad'.format(sys_ifname))
        return ipv6_set

    def update_interfaces_file(self):
        with open(self.INTERFACES_FILE, 'a') as f:
            f.write('source %s\n' % self.__get_cfg_fname())

    def _generate_keys(self):
        if self._l_private and self._l_public:
            return
        
        if os.path.exists('/etc/wg/private.key') and os.path.exists('/etc/wg/public.key'):
            with open('/etc/wg/private.key', 'r') as f:
                private_key = f.read()
            with open('/etc/wg/public.key', 'r') as f:
                public_key = f.read()
        else:
            if not os.path.exists('/etc/wg'):
                os.mkdir('/etc/wg/')
            # mkdir /etc/wg/
            vpp_if = vppif.VppApiInterface(self.LOG_NAME)
            ret, private_key, public_key = vpp_if.wg_generate_keys()
            del(vpp_if)
            
            if ret != 0:
                return

            with open('/etc/wg/private.key', 'w+') as private_file:
                private_file.write(str(private_key))
            with open('/etc/wg/public.key', 'w+') as public_file:
                public_file.write(str(public_key))

        self._l_private = private_key
        self._l_public = public_key
        self.logger.error("_generate_keys self._l_private {}, _l_public {}" .format(self._l_private, self._l_public))

    def is_ipv6(self):
        if self._l_endpoint_v6 and not self._l_endpoint:
            return True
        if self._l_endpoint_v6 and self._ip6_addr:
            return True
        return False

    def create_wg_interface(self):
        if self._wgif_idx:
            return
        if not self._l_port or not self._l_private or self._enable == False:
            return
        if self.is_ipv6 == True and not self._l_endpoint_v6:
            return
        if self.is_ipv6 == False and not self._l_endpoint:
            return

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        self.logger.error("wg_create_interface src_ip {} port {} private key {} - cfg_idx {}".format(self._l_endpoint_v6 if self.is_ipv6() == True else self._l_endpoint, self._l_port,self._l_private, int(self._name.replace('wg',''))))
        ret, ifidx = vpp_if.wg_create_interface(src_ip=self._l_endpoint_v6 if self.is_ipv6() == True else self._l_endpoint, port=self._l_port, private_key_base64=self._l_private, cfg_instance=int(self._name.replace('wg','')))
        self.logger.error("wg_create_interface ret {} ifidx {}".format(ret, ifidx))
        if ret != 0:
            self._wgif_idx = None
            del(vpp_if)
            return
            
        self._wgif_idx = ifidx
        vpp_if.ifup(self._wgif_idx)
        
        if self._mtu:
            ret = vpp_if.set_interface_mtu(self._wgif_idx, self._mtu)
        
        del(vpp_if)
        self._update_vpp_ip_address()
        self.add_peers()

    def delete_wg_interface(self):
        if not self._wgif_idx:
            return

        self.delete_peers()
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        ret = vpp_if.wg_delete_interface(sw_if_index=self._wgif_idx)
        del(vpp_if)
        self.logger.error("wg_delete_interface ret {}".format(ret))
        if ret == 0:
            self._wgif_idx = None
            for peer in self._peers:
                peer._peer_idx = None
            
    def _update_vpp_ip_address(self):
        if not self._wgif_idx:
            return
        
        if not self._ip4_addr and not self._ip6_addr:
            return
        
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        if self._tapidx:
            if self._ip4_addr:
                ret = vpp_if.drctl_set_ip(self._tapidx, self._ip4_addr.ip, delete=0 )
            if self._ip6_addr:
                ret = vpp_if.drctl_set_ip(self._tapidx, self._ip6_addr.ip, delete=0 )
        
        if self._ip4_addr:
            self.logger.error("WireGuardVPNInterface add address {}/{} to wireguard interface" .format(str(self._ip4_addr.ip), str(self._ip4_addr.network.netmask)))
            ret = vpp_if.add_interface_address(self._wgif_idx, '{}/{}' .format(str(self._ip4_addr.ip), str(self._ip4_addr.network.netmask)))
            if ret:
                self.logger.error("WireGuardVPNInterface error adding address to wireguard interface - {}" .format(ret))
        if self._ip6_addr:
            self.logger.error("WireGuardVPNInterface add address {} to wireguard interface" .format(str(self._ip6_addr.network)))
            ret = vpp_if.add_interface_address(self._wgif_idx, str(self._ip6_addr.network))
            ret = vpp_if.add_interface_address(self._wgif_idx, '{}/{}' .format(str(self._ip6_addr.ip), str(self._ip6_addr.network.prefixlen)))
            if ret:
                self.logger.error("WireGuardVPNInterface error adding address to wireguard interface")
        del(vpp_if)

    def add_del_peer_address_vpp_static_route(self, peer, delete=False):
        if not peer or not peer._allowed_ips or not self._wgif_idx:
            return

        netmask = 128 if peer._peer_is_ipv6 else 32
        if self._is_concentrator:
            self.logger.error("Routes for peers will be set from Wireguard Concentrator subscriber")
            return
        for a_ip in peer._allowed_ips:
            ip_addr = ipaddress.IPv6Interface(str(a_ip)) if peer._peer_is_ipv6 else ipaddress.IPv4Interface(str(a_ip))
            ip_netmask = ip_addr.network.prefixlen
            self.logger.error("WireGuardVPNInterface vpp_static_route ip_addr - {} ip_netmask {} - netmask {} - delete {}" .format(ip_addr, ip_netmask, netmask, delete))
            if netmask == ip_netmask:            
                vpp_if = vppif.VppApiInterface(self.LOG_NAME)
                if delete == True:
                    ret = vpp_if.del_ip_route('{}/{}' .format(ip_addr.ip, ip_netmask), next_hop_sw_if_index=self._wgif_idx)
                    if ret:
                        self.logger.error ('Del ip route fail route {}/{} via {} ret={}'.format(ip_addr.ip, ip_netmask, self._wgif_idx ,ret))
                    self.logger.error("WireGuardVPNInterface vpp_static_route static route for - {}/{} REMOVED" .format(ip_addr.ip, ip_netmask))
                else:
                    ret = vpp_if.add_ip_route('{}/{}' .format(ip_addr.ip, ip_netmask), next_hop_sw_if_index=self._wgif_idx, check=False)
                    if ret:
                        self.logger.error ('Add ip route fail route {}/{} via {} ret={}'.format(ip_addr.ip, ip_netmask, self._wgif_idx ,ret))
                    self.logger.error("WireGuardVPNInterface vpp_static_route static route for - {}/{} ADDED" .format(ip_addr.ip, ip_netmask))
                del(vpp_if)

    def add_del_peers_address_vpp_static_route(self, delete=False):
        if not self._wgif_idx or not self._peers:
            return
        
        for peer in self._peers:
            self.add_del_peer_address_vpp_static_route(peer, delete)

    def add_peers(self):
        if not self._wgif_idx or not self._peers:
            self.logger.error("WireGuardVPNInterface add peers return")
            return
        
        for peer in self._peers:
            self.add_peer(peer)

    def add_peer(self, peer):
        if not peer:
            return
        if not self._wgif_idx or not self._peers:
            return
        
        if not peer._r_public:
            peer._get_peer_public_key()
        if not peer._r_public or not peer._r_port or not peer._r_endpoint or not peer._allowed_ips:
            self.logger.error("WireGuardVPNInterface add peer return {} - {} - {} - {}" .format(peer._r_public, peer._r_port, peer._r_endpoint, peer._allowed_ips))
            return
        
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        ret, peer_idx = vpp_if.wg_add_peer(wgif=self._wgif_idx, endpoint=peer._r_endpoint, port=peer._r_port, keepalive=peer._persistent_keepalive,
                            allowed_ips=peer._allowed_ips, public_key_base64=peer._r_public, base64=True)
        del(vpp_if)
        self.logger.error("add_peers ret {} peer_idx {}".format(ret, peer_idx))
        if ret:
            self.logger.error("Error adding peer {}" .format(peer._r_endpoint))
            return
        peer._peer_idx = peer_idx
        
        self.add_del_peer_address_vpp_static_route(peer, delete=False)

    def delete_peers(self):

        # delete routes on concentrator from clients
        if self._is_concentrator:
            wgc.WireGuardConcentratorClient.WireGuardConcentratorClientDelRoutes(self.name, self._wgif_idx)
            self.logger.debug("WireGuardConcentratorClientDelRoutes called - routes deleted")

        if not self._wgif_idx or not self._peers:
            self.logger.error("WireGuardVPNInterface delete peers return")
            return
        
        for peer in self._peers:
            self.delete_peer(peer)

    def delete_peer(self, peer):
        if not peer or not self._wgif_idx:
            return
        if peer._peer_idx == None:
            return
        
        self.logger.error("WireGuardVPNInterface delete peer {}" .format(peer._peer_idx))
        self.add_del_peer_address_vpp_static_route(peer, delete=True)
        
        if self._conc_endpoint:
            self.del_routes_from_concentrator(peer)

        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        ret = vpp_if.wg_del_peer(peer_idx=peer._peer_idx)
        del(vpp_if)
        self.logger.error(" wg_del_peer idx {} return {}".format(peer._peer_idx, ret))
        if ret:
            self.logger.error("Error deleting peer {}" .format(peer._r_endpoint))
            return
        peer._peer_idx = None

    def _get_wireguard_data(self):
        token = WireGuardPeer.get_auth_token_tveapi_r_endpoint(False, self._conc_endpoint, self.logger)
        if not token:
            return None

        url = f"https://{self._conc_endpoint}:7373/v1/wireguard/concentrator-client"
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token}
        data = {
            "public-key": self._encrypt_public_key(self._l_public).decode("utf-8"),
            "hostname": self.get_hostname(),
            "password": self._conc_password,
            "concentrator-port": self._conc_port
        }
        try:
            response = requests.put(url, headers=headers, data=json.dumps(data), verify=False, timeout=10)
            response_data = response.json()

            if "error" in response_data:
                self.logger.error(f"WireGuardConcentrator response: {response_data['error']}")
                raise WireGuardIfErr(f"Client already connected to Wireguard Concentrator!", -1)

            return response_data
        except Exception as ex:
            self.logger.error(f"WireGuardConcentrator Exception {ex}")

        return None
    
    def _set_domain_lookup(self, d_look=False):
        self.logger.debug(f"hostname {type(self.get_hostname())} d_look {d_look}")
        token = WireGuardPeer.get_auth_token_tveapi_r_endpoint(False, self._conc_endpoint, self.logger)
        if not token:
            return None

        url = f"https://{self._conc_endpoint}:7373/v1/wireguard/domain-lookup"
        headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }
        data = {
            "hostname": self.get_hostname(),
            "domain-lookup": d_look
        }
        try:
            response = requests.post(url, headers=headers, data=json.dumps(data), verify=False, timeout=10)
            self.logger.error(f"WireGuardConcentrator response: {response.text}")            
            return response.json()
        except Exception as ex:
            self.logger.error(f"WireGuardConcentrator Exception {ex}")

        return None
    
    @staticmethod
    def get_hostname():
        return os.uname()[1]
    
    def _generate_key(self):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=b'1V0eG071nP!cH',
            iterations=100000,
            backend=default_backend()
        )
        key = base64.urlsafe_b64encode(kdf.derive(self._conc_password.encode()))
        return key

    def _decrypt_public_key(self):
        cipher = Fernet(self._generate_key())
        decrypted_public_key = cipher.decrypt(self._conc_pub_key.encode()).decode()
        self._conc_pub_key = decrypted_public_key.strip()
        if not self._conc_pub_key:
            self.logger.error(f"Could not get Wireguard Concentrator public key!")

    def _encrypt_public_key(self, public_key):
        cipher = Fernet(self._generate_key())
        encrypted_public_key = cipher.encrypt(public_key.encode())
        return encrypted_public_key

    def add_ip_route_kernel(self, destination, gateway=None, oif_index=None, onlink=False):
            with IPRoute() as ipr:
                if gateway:
                    for ip in destination:
                        try:
                            # Attempt to add the route
                            self.logger.error(f"IP: destination {destination} ip {ip} {type(ip)} gateway: {gateway}--------")
                            if onlink:
                                ipr.route("add", dst=ip, gateway=gateway, onlink=onlink)
                            else:
                                ipr.route("add", dst=ip, gateway=gateway)

                            self.logger.error("Route added successfully")
                        except Exception as ex:
                            # Ignore "File exists" error
                            if ex.args[0] == 17:  # Error code for 'File exists'
                                self.logger.error(f"Route already exists, ignoring: {ex}")
                            else:
                                self.logger.error(f"Failed to add route: {ex}")
                elif oif_index:
                    try:
                        ipr.route("add", dst=destination, oif=oif_index, scope='link')
                        self.logger.error("First route added")
                    except Exception as ex:
                        # Ignore "File exists" error
                        if ex.args[0] == 17:  # Error code for 'File exists'
                            self.logger.error(f"Route already exists, ignoring: {ex}")
                        else:
                            self.logger.error(f"Failed to add route: {ex}")
                
    def add_del_addresses_vpp_static_route(self, ip_address, next_hop_addr=None, delete=False):
        if not self._wgif_idx or not self._conc_endpoint:
            return
        
        ip_addr = ipaddress.IPv4Interface(str(ip_address))
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        if delete == True:
            ret = vpp_if.del_ip_route(ip_addr, next_hop_sw_if_index=self._wgif_idx, next_hop_address=next_hop_addr)
            if ret:
                self.logger.error ('Del ip route fail route {} via {} ret={}'.format(ip_addr, self._wgif_idx ,ret))
            self.logger.error("WireGuardVPNInterface vpp_static_route static route for - {} REMOVED" .format(ip_addr))
        else:
            ret = vpp_if.add_ip_route(ip_addr, next_hop_sw_if_index=self._wgif_idx, next_hop_address=next_hop_addr, check=False)
            if ret:
                self.logger.error ('Add ip route fail route {} via {} ret={}'.format(ip_addr, self._wgif_idx ,ret))
            self.logger.error("WireGuardVPNInterface vpp_static_route static route for - {} ADDED" .format(ip_addr))
        del(vpp_if)
        
    def del_routes_from_concentrator(self, peer):
        if peer._is_concentrator:
            self.add_del_addresses_vpp_static_route(f"{self._conc_tunnel_ip}/32", delete=True)
            for a_ip in self._client_allowed_ips.split(','):
                self.add_del_addresses_vpp_static_route(a_ip, next_hop_addr=self._conc_tunnel_ip, delete=True)

    def _add_concentrator_routes(self):
        if self.if_wg_concentrator_valid_prerequisites():
            with IPRoute() as ipr:
                oif = ipr.link_lookup(ifname=self._name)
            self.logger.error(f"OIF: {oif} conc_tun_ip: {self._conc_tunnel_ip} client_al_ips: {self._client_allowed_ips}")
            self.add_ip_route_kernel(f"{self._conc_tunnel_ip}/32", oif_index=oif[0])
            self.add_del_addresses_vpp_static_route(f"{self._conc_tunnel_ip}/32")
            self.add_ip_route_kernel(self._client_allowed_ips.split(','), gateway=self._conc_tunnel_ip, onlink=True)
            for a_ip in self._client_allowed_ips.split(','):
                self.add_del_addresses_vpp_static_route(a_ip, next_hop_addr=self._conc_tunnel_ip)
                
            os.system(f"/scripts/wg_client_thread.py {self._name} {self._conc_tunnel_ip} {self._l_public} &")

    def _set_domain_name(self, domain_name):
        self.replace_or_append_string_in_file(self.DNS_CONF_FILE, 'local=', f'local=/{domain_name}/')
        self.replace_or_append_string_in_file(self.DNS_CONF_FILE, 'domain=', f'domain={domain_name}')

    @staticmethod
    def replace_or_append_string_in_file(filename, str_to_replace, new_str):
        new_string = new_str + '\n'
        string_exists = False

        # Check if the string exists in the file
        with open(filename, "r") as f:
            for line in f:
                if line.strip().startswith(str_to_replace.strip()):
                    string_exists = True
                    break

        # If the string doesn't exist, append it to the file
        if not string_exists:
            with open(filename, "a") as f:
                f.write(new_string)
        else:
            # Replace the string in the file
            for line in fileinput.input([filename], inplace=True):
                if line.strip().startswith(str_to_replace):
                    line = new_string
                sys.stdout.write(line)

    def _create_dns_hosts_dir(self):
        if not os.path.exists(self.DNSMASQ_HOSTS_DIR):
            os.makedirs(self.DNSMASQ_HOSTS_DIR)
            os.system('systemctl restart dnsmasq.service')
            
    def _enable_dnsmasq(self):
        utils.execute_shell_command(f'sudo systemctl stop systemd-resolved.service')
        utils.execute_shell_command(f'sudo systemctl disable systemd-resolved.service')
        
        if os.path.exists('/etc/resolv.conf'):
            utils.execute_shell_command(f'sudo unlink /etc/resolv.conf')
        
        with open('/etc/resolv.conf', 'w') as f:
            print('nameserver 127.0.0.1', file=f)
        
        self._create_dns_hosts_dir()

        utils.execute_shell_command(f'sudo systemctl unmask dnsmasq.service')
        utils.execute_shell_command(f'sudo systemctl restart dnsmasq.service')
    
    def _set_conc_domain_locally(self):
        subprocess.call('/usr/bin/resolvectl dns {} {}'.format(self._name, self._conc_tunnel_ip), shell=True)
        subprocess.call('/usr/bin/resolvectl domain {} {}'.format(self._name, self._d_name), shell=True)
        subprocess.call('/usr/bin/resolvectl log-level 3', shell=True)

    def _dnsmasq_add_del_listen_interface(self, interface, delete=False):
        with open(self.DNS_CONF_FILE, 'r') as file:
            lines = file.readlines()

        interface_line_index = None
        for i, line in enumerate(lines):
            if line.startswith('interface='):
                interface_line_index = i
                break

        if interface_line_index is not None:
            current_interfaces = lines[interface_line_index].strip().split('=')[1].split(',')
            current_interfaces = [iface for iface in current_interfaces if iface]
        else:
            current_interfaces = []

        if not delete:
            if interface not in current_interfaces:
                current_interfaces.append(interface)
        else:
            if interface in current_interfaces:
                current_interfaces.remove(interface)

        if current_interfaces:
            new_interface_line = f"interface={','.join(current_interfaces)}\n"
        else:
            new_interface_line = "interface=\n"

        if interface_line_index is not None:
            # Replace the old interface line with the new one
            lines[interface_line_index] = new_interface_line
        else:
            # If no interface line was found, add a new one
            lines.append(new_interface_line)

        with open(self.DNS_CONF_FILE, 'w') as file:
            file.writelines(lines)

    def verify_if_table_exists_in_db(self, table_name):
        tables = self._edge_db.get_all_tables('wireguard')
        for table in tables:
            if table == table_name:
                return True

    def create_wireguard_keepalive_user(self):
        self._edge_db.create_user('wireguard', '1V0eG071nP!cH')
        self._edge_db.grant_privileges_to_user('wireguard', 'wireguard', 'timestamp', ['SELECT', 'INSERT', 'UPDATE'])
        self._edge_db.execute_and_close("GRANT SELECT, INSERT, UPDATE ON `wireguard`.`timestamp` TO `wireguard`@`%`;")
        self._edge_db.execute_and_close('FLUSH PRIVILEGES;')

    def get_pid_by_command(command_string):
        for proc in psutil.process_iter(['cmdline']):
            try:
                if proc.cmdline() == command_string.split():
                    return proc.pid
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                pass
        return None
    
    def _add_ip_tables_mariadb_rule(self):
        try:
            subprocess.check_output(
                ['/sbin/iptables', '-A', 'INPUT', '-p', 'tcp', '-m', 'tcp', '--dport', '3306', '-j', 'DROP'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Failed to add iptables rule! Exception: {e}")
    
    def _del_ip_tables_mariadb_rule(self):
        try:
            subprocess.check_output(
                ['/sbin/iptables', '-D', 'INPUT', '-p', 'tcp', '-m', 'tcp', '--dport', '3306', '-j', 'DROP'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Failed to del iptables rule! Exception: {e}")

    def set_host_mac_equal_to_tap_mac(self):
        with open(f"/sys/class/net/{self._name}/address") as file:
            host_mac = file.readline().split('\n')[0]
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        tap_mac = vpp_if.get_tap_interface_by_host_ifname(self._name)['host_mac_addr']
        if host_mac != tap_mac:
            subprocess.call(['ifconfig', self._name, 'hw', 'ether', f"{tap_mac}"])

    def __get_ip_from_hostname(self, conc_endpoint):
        try:
            if conc_endpoint:
                ip = socket.gethostbyname(conc_endpoint)
                return ip
        except Exception as e:
            self.logger.error(f'Invalid hostname, error raised is {e}')

    # Search for the exact ACL using the VPP API, remove it from any interface if it's applied and delete it.
    def _search_and_delete_acl(self, acl_json):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        acl_dump = vpp_if.acl_dump()

        # find the matching ACL
        for acl in acl_dump:
            for rule in acl['rules']:
                if (rule['src_prefix'] == IPv4Network(acl_json['src_prefix']) and
                    rule['dst_prefix'] == IPv4Network(acl_json['dst_prefix']) and
                    rule['is_permit'] == (1 if acl_json['action'] == 'permit' else 0)):

                    acl_index = acl['acl_index']
                    self.logger.error(f"acl_index {acl_index}")
                    # Check if the ACL is applied to any interface
                    acl_interfaces = vpp_if.acl_interface_list_dump()
                    for interface in acl_interfaces:
                        if acl_index in interface['acls']:
                            # Remove the ACL from the interface
                            input_acls = [acl for acl in interface['acls'] if acl != acl_index]
                            vpp_if.acl_interface_set_acl_list(
                                sw_if_index=interface['sw_if_index'],
                                input_acls=input_acls,
                                output_acls=[]  # We currently set only input ACLs
                            )
                            self.logger.error(f"Removed ACL {acl_index} from interface {interface['sw_if_index']}")

                    # Delete the ACL
                    ret = self._del_acl(acl_index)
                    self.logger.error(f"ACL {acl_index} deleted. ret {ret}")
                    return True

        self.logger.debug("ACL not found in VPP.")
        return False

    def _search_for_acl_by_tag(self, tag):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)
        acl_dump = vpp_if.acl_dump()

        matching_acls = []

        for acl in acl_dump:
            if acl.get('tag') == tag:
                matching_acls.append(acl['acl_index'])

        return matching_acls

    def _set_acl(self, intf_name, rules):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        # Ensure rules is a list, even if it's a single dictionary
        if isinstance(rules, dict):
            rules = [rules]

        acl_index = vpp_if.acl_add_replace(
                    tag=intf_name,
                    rules=rules
        )

        return acl_index

    def _set_bidirectional_acl(self, intf_name, acl_json):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        rules = [
            {'src_prefix': acl_json['src_prefix'], 'dst_prefix': acl_json['dst_prefix'], 'action': 'permit'},
            {'src_prefix': acl_json['dst_prefix'], 'dst_prefix': acl_json['src_prefix'], 'action': 'permit'}
        ]

        acl_index = vpp_if.acl_add_replace(
            tag=intf_name,
            rules=rules
        )

        return acl_index

    def _set_default_acl_rule(self, intf_name, conc_ip):
        acl_json = {
            'src_prefix': f'{conc_ip}/32',
            'dst_prefix': '0.0.0.0/0',
            'action': 'allow'
        }

        default_rule_idx = self._set_bidirectional_acl(intf_name, acl_json)
        self._default_acl_rule_idx = True

        return default_rule_idx

    def _del_acl(self, acl_index):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        ret = vpp_if.acl_del(acl_idx=acl_index)
        self.logger.debug(f"ACL {acl_index} deleted. ret {ret}")

    def _apply_acl_to_interface(self, sw_if_idx, acl_index):
        vpp_if = vppif.VppApiInterface(self.LOG_NAME)

        input_acls = acl_index
        output_acls = []

        result = vpp_if.acl_interface_set_acl_list(
                sw_if_index=sw_if_idx,
                input_acls=input_acls,
                output_acls=output_acls
        )

        return result

    def _process_acl_rules(self, new):
        # Ensure default acl rule can be created before any other rule
        if not self._ip4_addr:
            return
        for action, rules in new.items():
            # Determine if 'rules' is a single rule string or a list of rule dictionaries
            if isinstance(rules, str):
                rules = [eval(rules.replace("'", "\""))]  # Convert to list with a single dictionary

            elif isinstance(rules, list):
                rules = [rule for rule in rules]  # Already in correct format, but ensuring it's processed as a list

            for rule in rules:
                # Convert each rule to JSON format
                acl_json = self._acl_dict_to_json({action: rule})

                # Apply the ACL rule
                if action == 'permit':
                    if not self._default_acl_rule_idx:
                        self._set_default_acl_rule(self._name, self._ip4_addr.ip)
                    # Set ACL in both directions if the action is "permit"
                    self._set_bidirectional_acl(self._name, acl_json)
                else:
                    self._set_acl(self._name, acl_json)

                # Apply the ACL to the interface if required
                if self._is_concentrator and self._wgif_idx:
                    self._apply_acl_to_interface(self._wgif_idx, self._search_for_acl_by_tag(self._name))
                    self.logger.error(f"Access list {new} applied to {self._name}!")

    def _process_acl_deletion(self, acl):
        for action, rules_str in acl.items():
            if 'delete' in rules_str:
                # Remove ' delete' and convert to dictionary
                clean_rules_str = rules_str.replace(' delete', '')
                rule_data = eval(clean_rules_str.replace("'", "\""))  # Convert string to dict

                acl_dict = {action: rule_data}
                acl_json = self._acl_dict_to_json(acl_dict)
                self._search_and_delete_acl(acl_json)
                self.logger.error(f"Access list {acl} deleted from {self._name}!")
                return True

        return False

    def _acl_dict_to_json(self, acl_dict):
        for action, rule_data in acl_dict.items():
            if isinstance(rule_data, dict):
                # Process the dictionary directly
                rule = rule_data.get('rule', '')

                if rule:
                    src_dst = rule.split(' ')

                    if len(src_dst) == 4:
                        src_prefix = src_dst[1]
                        dst_prefix = src_dst[3]

                        acl_json = {
                            'src_prefix': src_prefix,
                            'dst_prefix': dst_prefix,
                            'action': action.lower()
                        }

                        return acl_json
            else:
                raise WireGuardIfErr("Invalid ACL dictionary format!", -1)

